#!/usr/bin/env python3
#
# Run with pytest

from typing import Optional, List
from pathlib import Path
from dataclasses import dataclass
import xml.etree

import os
import logging
import string
import pytest

from . import WacomDatabase, WacomDevice
from .conftest import load_test_db

logger = logging.getLogger(__name__)


def datadir():
    return Path(os.getenv("MESON_SOURCE_ROOT") or ".") / "data"


def layoutsdir():
    return datadir() / "layouts"


def load_svg(layoutfile: Path) -> xml.etree.ElementTree:
    try:
        f = layoutsdir() / layoutfile
        assert f.exists()
        tree = xml.etree.ElementTree.parse(f)
        assert tree is not None
        return tree
    except KeyError:
        pass
    return None


@dataclass
class SvgDevice:
    _db: WacomDatabase
    device: WacomDevice
    svg: xml.etree.ElementTree
    is_autogenerated: bool

    def has_item(self, id: str, classes: Optional[List[str]] = None):
        root = self.svg.getroot()
        nodes = root.findall(f".//*[@id='{id}']")
        assert nodes, f"Failed to find required element with id {id}"
        assert len(nodes) == 1, f"Expected one element with id {id}, have {len(nodes)}"
        node = nodes[0]
        for klass in classes or []:
            assert klass in node.get("class", "").split(" "), (
                f"Missing class '{klass}' for {id}. Have: {node.get('class')}"
            )

    @property
    def name(self):
        return self.device.name


def is_autogenerated(svg_filename: Path):
    tabletfile = datadir() / svg_filename.name.replace(".svg", ".tablet")
    return (
        any("autogenerated" in t for t in open(tabletfile))
        if tabletfile.exists()
        else False
    )


def pytest_generate_tests(metafunc):
    # We need to keep the db inside each SvgDevice, otherwise python may clean it up
    # and thus invalidate all our devices from db.list_devices()
    db = load_test_db()
    devices = db.list_devices()
    devices = [
        SvgDevice(
            db,
            d,
            load_svg(d.layout_filename),
            is_autogenerated(Path(d.layout_filename)),
        )
        for d in devices
        if d.layout_filename
    ]
    devices = sorted(devices, key=lambda d: d.name)

    def filenames(devices: List[SvgDevice]) -> List[str]:
        return [Path(d.device.layout_filename).name for d in devices]

    if "svgdevice" in metafunc.fixturenames:
        metafunc.parametrize("svgdevice", devices, ids=filenames(devices))

    if "ringdevice" in metafunc.fixturenames:
        devices = list(filter(lambda d: d.device.num_rings > 0, devices))
        metafunc.parametrize("ringdevice", devices, ids=filenames(devices))

    if "stripdevice" in metafunc.fixturenames:
        devices = list(filter(lambda d: d.device.num_strips > 0, devices))
        metafunc.parametrize("stripdevice", devices, ids=filenames(devices))

    if "dialdevice" in metafunc.fixturenames:
        devices = list(filter(lambda d: d.device.num_dials > 0, devices))
        metafunc.parametrize("dialdevice", devices, ids=filenames(devices))

    if "buttondevice" in metafunc.fixturenames:
        devices = list(filter(lambda d: d.device.num_buttons > 0, devices))
        metafunc.parametrize("buttondevice", devices, ids=filenames(devices))


def test_svg(svgdevice):
    root = svgdevice.svg.getroot()
    assert root.tag in ["svg", "{http://www.w3.org/2000/svg}svg"]
    assert root.get("width") is not None
    assert root.get("height") is not None


def test_svg_maybe_not_needed(svgdevice):
    device = svgdevice.device
    features = [
        device.num_buttons,
        device.num_rings,
        device.num_strips,
        device.num_dials,
    ]
    assert any(x > 0 for x in features), (
        f"Device {device.name} has no buttons/rings/strips/dials and should not have an SVG"
    )


def has_item(root, id: str, classes: Optional[List[str]] = None):
    nodes = root.findall(f".//*[@id='{id}']")
    assert nodes, f"Failed to find required element with id {id}"
    assert len(nodes) == 1, f"Expected on element with id {id}, have {len(nodes)}"
    node = nodes[0]
    for klass in classes or []:
        assert klass in node.get("class").split(" "), (
            f"Missing class '{klass}' for {id}. Have: {node.get('class')}"
        )


def test_svg_rings(ringdevice):
    if ringdevice.device.num_rings >= 1:
        ringdevice.has_item(id="Ring", classes=["Ring", "TouchRing"])
        ringdevice.has_item(id="LabelRingCW", classes=["RingCW", "Ring", "Label"])
        ringdevice.has_item(id="LabelRingCCW", classes=["RingCCW", "Ring", "Label"])
        ringdevice.has_item(id="LeaderRingCW", classes=["RingCW", "Ring", "Leader"])
        ringdevice.has_item(id="LeaderRingCCW", classes=["RingCCW", "Ring", "Leader"])

    if ringdevice.device.num_rings >= 2:
        ringdevice.has_item(id="Ring2", classes=["Ring2", "TouchRing"])
        ringdevice.has_item(id="LabelRing2CW", classes=["Ring2CW", "Ring2", "Label"])
        ringdevice.has_item(id="LabelRing2CCW", classes=["Ring2CCW", "Ring2", "Label"])
        ringdevice.has_item(id="LeaderRing2CW", classes=["Ring2CW", "Ring2", "Leader"])
        ringdevice.has_item(
            id="LeaderRing2CCW", classes=["Ring2CCW", "Ring2", "Leader"]
        )


def test_svg_strips(stripdevice):
    try:
        if stripdevice.device.num_strips >= 1:
            stripdevice.has_item(id="Strip", classes=["Strip", "TouchStrip"])
            stripdevice.has_item(
                id="LabelStripUp", classes=["StripUp", "Strip", "Label"]
            )
            stripdevice.has_item(
                id="LabelStripDown", classes=["StripDown", "Strip", "Label"]
            )
            stripdevice.has_item(
                id="LeaderStripUp", classes=["StripUp", "Strip", "Leader"]
            )
            stripdevice.has_item(
                id="LeaderStripDown", classes=["StripDown", "Strip", "Leader"]
            )

        if stripdevice.device.num_strips >= 2:
            stripdevice.has_item(id="Strip2", classes=["Strip2", "TouchStrip"])
            stripdevice.has_item(
                id="LabelStrip2Up", classes=["Strip2Up", "Strip2", "Label"]
            )
            stripdevice.has_item(
                id="LabelStrip2Down", classes=["Strip2Down", "Strip2", "Label"]
            )
            stripdevice.has_item(
                id="LeaderStrip2Up", classes=["Strip2Up", "Strip2", "Leader"]
            )
            stripdevice.has_item(
                id="LeaderStrip2Down", classes=["Strip2Down", "Strip2", "Leader"]
            )
    except AssertionError as e:
        if stripdevice.is_autogenerated:
            pytest.skip(f"Autogenerated device has errors in SVG: {e}")
        raise e


def test_svg_dials(dialdevice):
    try:
        if dialdevice.device.num_dials >= 1:
            dialdevice.has_item(id="Dial", classes=["Dial", "TouchDial"])
            dialdevice.has_item(id="LabelDialCW", classes=["DialCW", "Dial", "Label"])
            dialdevice.has_item(id="LabelDialCCW", classes=["DialCCW", "Dial", "Label"])
            dialdevice.has_item(id="LeaderDialCW", classes=["DialCW", "Dial", "Leader"])
            dialdevice.has_item(
                id="LeaderDialCCW", classes=["DialCCW", "Dial", "Leader"]
            )

        if dialdevice.device.num_dials >= 2:
            dialdevice.has_item(id="Dial2", classes=["Dial2", "TouchDial"])
            dialdevice.has_item(
                id="LabelDial2CW", classes=["Dial2CW", "Dial2", "Label"]
            )
            dialdevice.has_item(
                id="LabelDial2CCW", classes=["Dial2CCW", "Dial2", "Label"]
            )
            dialdevice.has_item(
                id="LeaderDial2CW", classes=["Dial2CW", "Dial2", "Leader"]
            )
            dialdevice.has_item(
                id="LeaderDial2CCW", classes=["Dial2CCW", "Dial2", "Leader"]
            )
    except AssertionError as e:
        if dialdevice.is_autogenerated:
            pytest.skip(f"Autogenerated device has errors in SVG: {e}")
        raise e


def test_svg_button(buttondevice):
    for button in list(string.ascii_uppercase)[: buttondevice.device.num_buttons]:
        modeswitch_label = []
        flags = buttondevice.device.button_flags(button)
        if any(f in flags for f in WacomDevice.ButtonFlags.modeswitch_flags()):
            modeswitch_label = ["ModeSwitch"]
        buttondevice.has_item(
            id=f"Button{button}", classes=["Button", button] + modeswitch_label
        )
        buttondevice.has_item(
            id=f"Label{button}", classes=["Label", button] + modeswitch_label
        )
        buttondevice.has_item(
            id=f"Leader{button}", classes=["Leader", button] + modeswitch_label
        )



#! /usr/bin/env python3
#
# Copyright (c) 2013 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# Author: Joaquim Rocha <jrocha@redhat.com>
#

import logging
import sys
from pathlib import Path
from argparse import ArgumentParser
from xml.etree import ElementTree as ET

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("clean-svg")
logger.setLevel(logging.INFO)

NAMESPACE = "http://www.w3.org/2000/svg"
BRACKETS_NAMESPACE = "{" + NAMESPACE + "}"


def human_round(number):
    """
    Round to closest .5, keep integer values
    """
    v = round(number * 2) / 2.0
    return int(v) if v == int(v) else v


def traverse_and_clean(node):
    """
    Clean the tree recursively
    """
    # Remove any non-SVG namespace attributes
    for key in list(node.attrib.keys()):
        if key.startswith("{"):
            del node.attrib[key]
    if "style" in node.attrib:
        if node.tag == "text":
            node.attrib["style"] = "text-anchor:start;"
        elif node.tag != "svg":
            del node.attrib["style"]

    remove_transform_if_exists(node)

    round_attrib(node, "d", "x", "y", "rx", "ry", "width", "height", "cx", "cy", "r")

    for child in node:
        traverse_and_clean(child)


def round_attrib(node, *attrs):
    for attr_name in attrs:
        attr_value = node.attrib.get(attr_name)
        if attr_value is None:
            continue
        if attr_name == "d":
            d = attr_value.replace(",", " ")
            values = [round_if_number(value) for value in d.split()]
            node.attrib[attr_name] = " ".join(values)
        else:
            node.attrib[attr_name] = round_if_number(attr_value)


def round_if_number(value):
    try:
        value = human_round(float(value.strip()))
    except ValueError:
        pass
    return str(value)


def remove_non_svg_nodes_and_strip_namespace(root):
    if root.tag.startswith(BRACKETS_NAMESPACE):
        root.tag = root.tag[len(BRACKETS_NAMESPACE) :]
    for elem in list(root):
        if (
            not elem.tag.startswith(BRACKETS_NAMESPACE)
            or elem.tag == BRACKETS_NAMESPACE + "metadata"
        ):
            root.remove(elem)
        else:
            remove_non_svg_nodes_and_strip_namespace(elem)


def remove_transform_if_exists(node):
    TRANSLATE = "translate"
    MATRIX = "matrix"

    transform = node.attrib.get("transform")
    if transform is None:
        return
    transform = transform.strip()

    if transform.startswith(TRANSLATE):
        values = transform[len(TRANSLATE) + 1 : -1].split(",")
        try:
            x, y = float(values[0]), float(values[1])
        except Exception:
            return

        apply_translation(node, 1, 0, 0, 1, x, y)
    elif transform.startswith(MATRIX):
        values = transform[len(MATRIX) + 1 : -1].split(",")
        try:
            a, b, c, d, e, f = [float(value.strip()) for value in values]
        except Exception:
            return
        apply_translation(node, a, b, c, d, e, f)
        apply_scaling(node, a, d)
    del node.attrib["transform"]


def apply_translation(node, a, b, c, d, e, f):
    x_attr, y_attr = "x", "y"
    if node.tag == "circle":
        x_attr, y_attr = "cx", "cy"
    elif node.tag == "path":
        apply_translation_to_path(node, e, f)
        return
    try:
        x, y = float(node.attrib[x_attr]), float(node.attrib[y_attr])
        new_x = x * a + y * c + 1 * e
        new_y = x * b + y * d + 1 * f
        node.attrib[x_attr] = str(new_x)
        node.attrib[y_attr] = str(new_y)
    except Exception:
        pass


def apply_translation_to_path(node, x, y):
    d = node.attrib.get("d")
    if d is None:
        return
    d = d.replace(",", " ").split()
    m_init_index = 0
    length = len(d)
    m_end_index = length
    operation = "M"
    i = 0
    while i < length:
        value = d[i]
        if value.lower() == "m":
            operation = value
            m_init_index = i + 1
        elif len(value) == 1 and value.isalpha():
            m_end_index = i
            break
        i += 1
    for i in range(m_init_index, m_end_index, 2):
        d[i] = str(float(d[i]) + x)
        d[i + 1] = str(float(d[i + 1]) + y)
        if operation == "m":
            break
    node.attrib["d"] = (
        " ".join(d[:m_init_index])
        + " "
        + " ".join(d[m_init_index:m_end_index])
        + " ".join(d[m_end_index:])
    )


def apply_scaling(node, x, y):
    w_attr, h_attr = "width", "height"
    if node.tag == "circle":
        r = float(node.attrib.get("r", 1.0))
        node.attrib["r"] = str(r * x)
    try:
        w = float(node.attrib[w_attr])
        h = float(node.attrib[h_attr])
        node.attrib[w_attr] = str(w * x)
        node.attrib[h_attr] = str(h * y)
    except Exception:
        pass


def to_string_rec(node, level=0):
    indent = "\n" + level * "  "

    tag_name = node.tag

    # Remove 'defs' element. This cannot be done in the traverse_and_clean
    # because somehow it is never found
    if tag_name == "defs":
        return ""

    # use a list to put id and class as the first arguments
    attribs = []
    for attr in get_node_attrs_sorted(node):
        attr_value = node.attrib.get(attr)
        if attr_value is not None:
            attribs.append(indent + '   %s="%s"' % (attr, attr_value))

    string = indent + "<" + tag_name + "".join(attribs)
    if len(node) or node.text:
        string += ">"
        if not node.text or not node.text.strip():
            node.text = indent + "  "
        else:
            string += node.text
        if list(node):
            for child in get_node_children_sorted(node):
                string += to_string_rec(child, level + 1)
            string += indent
        string += "</%s>" % tag_name
    else:
        string += " />"
    return string


def custom_tag_sort(arg):
    """
    Use as key functon in sorted().

    Pre-fix arg tag name by a number in the sort order we want. Anything
    unspecified defaults to 9. i.e. circle -> 1circle, thus sorts lower
    than
    other tags.
    """
    tag_order = {"title": 0, "rect": 1, "circle": 2, "path": 3}
    return str(tag_order.get(arg.tag, 9)) + arg.tag


def get_node_children_sorted(node):
    return sorted(node, key=custom_tag_sort)


def custom_attr_sort(arg):
    """
    Same as custom_tag_sort but for a node's attributes
    """
    attr_order = {
        "id": 0,
        "class": 1,
        "x": 2,
        "y": 3,
        "cx": 4,
        "cy": 5,
        "width": 6,
        "height": 7,
    }
    return str(attr_order.get(arg, 9)) + arg


def get_node_attrs_sorted(node):
    attrs = node.attrib.keys()
    return sorted(attrs, key=custom_attr_sort)


def apply_id_and_class_from_group(group_node):
    button_assigned = label_assigned = path_assigned = False
    _id = group_node.attrib.get("id")
    if _id is None:
        return
    for child in group_node:
        if child.tag == "rect" or child.tag == "circle":
            if button_assigned:
                continue
            child.attrib["id"] = "Button%s" % _id
            child.attrib["class"] = "%s Button" % _id
            button_assigned = True
        elif child.tag == "path":
            if path_assigned:
                continue
            child.attrib["id"] = "Leader%s" % _id
            child.attrib["class"] = "%s Leader" % _id
            path_assigned = True
        elif child.tag == "text":
            if label_assigned:
                continue
            child.attrib["id"] = "Label%s" % _id
            child.attrib["class"] = "%s Label" % _id
            child.text = _id
            label_assigned = True


def to_string(root):
    header = """<?xml version="1.0" standalone="no"?>
<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"
   "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">"""
    return header + to_string_rec(root)


def clean_svg(root, tabletname):
    remove_non_svg_nodes_and_strip_namespace(root)
    title = root.find("title")
    if title is not None:
        title.text = tabletname
    root.attrib["xmlns"] = "http://www.w3.org/2000/svg"
    traverse_and_clean(root)


if __name__ == "__main__":
    parser = ArgumentParser(description="Clean SVG files for libwacom")
    parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Print debug logging output",
    )
    parser.add_argument(
        "--ignore-missing",
        action="store_true",
        default=False,
        help="Ignore .tablet files without a Layout",
    )
    parser.add_argument(
        "filename",
        type=str,
        help="SVG file to clean or .tablet file with a Layout=xyz.svg line",
        metavar="FILE",
    )
    parser.add_argument(
        "tabletname",
        type=str,
        nargs="?",
        help="The name of the tablet, can be omitted if FILE is a .tablet file",
        metavar="TABLET_NAME",
    )
    args = parser.parse_args()

    if args.debug:
        logger.setLevel(logging.DEBUG)

    svgfile = args.filename
    tabletname = args.tabletname
    if args.filename.endswith(".tablet"):
        import configparser

        config = configparser.ConfigParser()
        config.read(args.filename)
        try:
            svgname = config["Device"]["Layout"]
            if not svgname:
                raise KeyError("Empty layout line")
        except KeyError:
            print(
                f"{args.filename} does not specify a layout, skipping", file=sys.stderr
            )
            sys.exit(0 if args.ignore_missing else 77)
        svgfile = Path(args.filename).parent / "layouts" / svgname
        tabletname = config["Device"]["Name"]
    else:
        assert tabletname, "Argument TABLET_NAME must be provided for SVG files"

    ET.register_namespace("", NAMESPACE)
    try:
        tree = ET.parse(svgfile)
    except Exception as e:
        sys.stderr.write(str(e) + "\n")
        sys.exit(1)
    root = tree.getroot()
    clean_svg(root, tabletname)
    print(to_string(root))



#!/usr/bin/env python3
#
# This file is formatted with ruff format

from configparser import ConfigParser
from dataclasses import dataclass, field

import ctypes
import logging
import pytest
import string

from . import (
    WacomAxisType,
    WacomBuilder,
    WacomBustype,
    WacomDatabase,
    WacomDevice,
    WacomEraserType,
    WacomStatusLed,
    WacomStylus,
    WacomStylusType,
)

logger = logging.getLogger(__name__)


@dataclass
class StylusEntry:
    id: str
    name: str
    group: str = "generic-with-eraser"
    paired_stylus_ids: list[str] = field(default_factory=list)
    buttons: int = 2
    axes: str = "Tilt;Pressure;Distance;"
    stylus_type: str = "General"
    eraser_type: str | None = None
    alias_of: str | None = None
    has_wheel: str | None = None
    has_lens: str | None = None

    def add_to_config(self, config: ConfigParser):
        c = {
            "Name": self.name,
            "Group": self.group,
            "PairedStylusIds": ";".join(self.paired_stylus_ids),
            "Buttons": f"{self.buttons}",
            "Axes": self.axes,
            "Type": self.stylus_type,
        }
        for attr in ["eraser_type", "alias_of", "has_wheel", "has_lens"]:
            if (val := getattr(self, attr, None)) is not None:
                key = attr.title().replace("_", "")
                # We need lowercase true/false
                c[key] = val if not isinstance(val, bool) else str(val).lower()

        config[self.id] = c

    @classmethod
    def generic_pen(cls) -> "StylusEntry":
        return cls(
            id="0x0:0xaffff", name="General Pen", paired_stylus_ids=["0x0:0xafffe"]
        )

    @classmethod
    def generic_eraser(cls) -> "StylusEntry":
        return StylusEntry(
            id="0x0:0xafffe",
            name="General Pen Eraser",
            paired_stylus_ids=["0x0:0xaffff"],
            eraser_type="Invert",
        )


@dataclass
class StylusFile:
    entries: list[StylusEntry]

    @classmethod
    def default(cls) -> "StylusFile":
        return cls(
            entries=[
                StylusEntry.generic_pen(),
                StylusEntry.generic_eraser(),
            ]
        )

    def write_to_dir(self, dir, filename="libwacom.stylus"):
        config = ConfigParser()
        config.optionxform = lambda option: option

        for s in self.entries:
            s.add_to_config(config)

        with open(dir / filename, "w") as fd:
            config.write(fd, space_around_delimiters=False)

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug(f"{dir}/{filename}:")
            for line in open(dir / filename).readlines():
                logger.debug(f"  {line.rstrip()}")


@dataclass
class TabletFile:
    name: str
    matches: list[str]
    width: int = 229
    height: int = 152
    integrated_in: str = ""
    klass: str = "Bamboo"
    layout: str = ""
    has_stylus: bool = True
    is_reversible: bool = False
    styli: list[str] = field(default_factory=list)
    # extra additions to the tablet file, e.g. { "Buttons": { "Left" : "A;B;" }
    extra: dict[str, dict[str, str]] = field(default_factory=dict)

    def write_to(self, filename):
        config = ConfigParser()
        config.optionxform = lambda option: option

        cfg = {}
        cfg["Device"] = {
            "Name": self.name,
            "DeviceMatch": ";".join(self.matches),
            "Width": self.width,
            "Height": self.height,
            "IntegratedIn": self.integrated_in,
            "Class": self.klass,
            "Layout": self.layout,
        }
        if self.styli:
            cfg["Device"]["Styli"] = ";".join(self.styli)

        cfg["Features"] = {
            "Stylus": self.has_stylus,
            "Reversible": self.is_reversible,
        }

        cfg = cfg | self.extra

        for k, v in cfg.items():
            config[k] = v

        with open(filename, "w") as fd:
            config.write(fd, space_around_delimiters=False)

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug(f"{filename}:")
            for line in open(filename).readlines():
                logger.debug(f"  {line.rstrip()}")


@pytest.fixture()
def custom_datadir(tmp_path):
    StylusFile.default().write_to_dir(tmp_path)
    TabletFile(name="Generic", matches=["generic"]).write_to(
        tmp_path / "generic.tablet"
    )
    return tmp_path


def test_database_init(db):
    """Just a test to make sure it doesn't crash"""
    assert db is not None


def test_invalid_device(db):
    device = db.new_from_usbid(0x0, 0x0)
    assert device is None


def test_intuos4(db):
    device = db.new_from_usbid(0x056A, 0x00BC)
    assert device is not None

    assert device.name == "Wacom Intuos4 WL"
    assert device.get_class() == device.CLASS_INTUOS4
    assert device.vendor_id == 0x56A
    assert device.product_id == 0xBC
    assert device.bustype == WacomBustype.USB
    assert device.num_buttons == 9
    assert device.has_stylus()
    assert device.is_reversible()
    assert not device.has_touch()
    assert device.has_ring()
    assert not device.has_ring2()
    assert device.num_rings == 1
    assert not device.has_touchswitch()
    assert device.num_strips == 0
    assert device.num_dials == 0
    assert device.integration_flags == []
    assert device.width == 8
    assert device.height == 5

    matches = device.matches
    assert len(matches) == 2
    assert any(match.bustype == device.bustype for match in matches)
    assert any(match.vendor_id == device.vendor_id for match in matches)
    assert any(match.product_id == device.product_id for match in matches)


def test_intuos4_wl(db):
    device = db.new_from_usbid(0x056A, 0x00B9)
    assert device is not None

    assert WacomDevice.ButtonFlags.RING_MODESWITCH in device.button_flags("A")
    assert WacomDevice.ButtonFlags.OLED in device.button_flags("I")
    assert device.ring_num_modes == 4


def test_cintiq24hd(db):
    device = db.new_from_usbid(0x056A, 0x00F4)
    assert device is not None

    assert device.ring_num_modes == 3
    assert device.ring2_num_modes == 3


def test_cintiq21ux(db):
    device = db.new_from_usbid(0x056A, 0x00CC)
    assert device is not None

    assert device.num_strips == 2
    assert device.num_dials == 0


def test_wacf004(db):
    device = db.new_from_name("Wacom Serial Tablet WACf004")
    assert device is not None
    assert device.model_name is None
    assert device.integration_flags == [
        WacomDevice.IntegrationFlags.DISPLAY,
        WacomDevice.IntegrationFlags.SYSTEM,
    ]


def test_cintiq24hdt(db):
    device = db.new_from_usbid(0x056A, 0x00F8)
    assert device is not None

    match = device.paired_device
    assert match is not None
    assert match.vendor_id == 0x56A
    assert match.product_id == 0xF6
    assert match.bustype == device.BUSTYPE_USB

    modes = {"A": 0, "B": 1, "C": 2, "I": 0, "J": 1, "K": 2}
    for btn in string.ascii_uppercase[: device.num_buttons]:
        expected_mode = modes.get(btn, WacomDevice.ModeSwitch.NEXT)
        assert device.button_modeswitch_mode(btn) == expected_mode


def test_cintiq13hd(db):
    libevdev = pytest.importorskip("libevdev")
    device = db.new_from_name("Wacom Cintiq 13HD")
    assert device is not None

    assert device.button_evdev_code("A") == libevdev.EV_KEY.BTN_0.value
    assert device.button_evdev_code("B") == libevdev.EV_KEY.BTN_1.value
    assert device.button_evdev_code("C") == libevdev.EV_KEY.BTN_2.value
    assert device.button_evdev_code("D") == libevdev.EV_KEY.BTN_3.value
    assert device.button_evdev_code("E") == libevdev.EV_KEY.BTN_4.value
    assert device.button_evdev_code("F") == libevdev.EV_KEY.BTN_5.value
    assert device.button_evdev_code("G") == libevdev.EV_KEY.BTN_6.value
    assert device.button_evdev_code("H") == libevdev.EV_KEY.BTN_7.value
    assert device.button_evdev_code("I") == libevdev.EV_KEY.BTN_8.value
    assert device.model_name == "DTK-1300"


def test_cintiqpro13(db):
    device = db.new_from_name("Wacom Cintiq Pro 13")
    assert device is not None
    assert device.num_keys == 5

    for btn in string.ascii_uppercase[: device.num_buttons]:
        assert device.button_modeswitch_mode(btn) == WacomDevice.ModeSwitch.NEXT


def test_dell_canvas(db):
    device = db.new_from_name("Dell Canvas 27")
    assert device is not None
    assert device.integration_flags == [WacomDevice.IntegrationFlags.DISPLAY]


def test_bamboo_pen(db):
    libevdev = pytest.importorskip("libevdev")

    device = db.new_from_name("Wacom Bamboo Pen")
    assert device is not None
    assert device.button_evdev_code("A") == libevdev.EV_KEY.BTN_BACK.value
    assert device.button_evdev_code("B") == libevdev.EV_KEY.BTN_FORWARD.value
    assert device.button_evdev_code("C") == libevdev.EV_KEY.BTN_LEFT.value
    assert device.button_evdev_code("D") == libevdev.EV_KEY.BTN_RIGHT.value
    assert device.model_name == "MTE-450"


def test_isdv4_4800(db):
    device = db.new_from_usbid(0x56A, 0x4800)
    assert device is not None

    assert device.integration_flags == [
        WacomDevice.IntegrationFlags.DISPLAY,
        WacomDevice.IntegrationFlags.SYSTEM,
    ]
    assert device.model_name is None

    assert device.vendor_id == 0x56A
    assert device.product_id == 0x4800
    assert device.num_buttons == 0


def test_mobilestudio_pro_modeswitch(db):
    device = db.new_from_name("Wacom MobileStudio Pro 13")
    assert device is not None

    modes = {"H": 0, "I": 1, "J": 2, "K": 3}
    for btn in string.ascii_uppercase[: device.num_buttons]:
        expected_mode = modes.get(btn, WacomDevice.ModeSwitch.NEXT)
        assert device.button_modeswitch_mode(btn) == expected_mode


@pytest.mark.parametrize(
    "usbid,expected",
    [
        [(0x256C, 0x0067), [WacomStylus.Generic.PEN_NO_ERASER]],
        [
            (0x04F3, 0x264C),
            [
                WacomStylus.Generic.PEN_WITH_ERASER,
                WacomStylus.Generic.ERASER,
                WacomStylus.Generic.PEN_NO_ERASER,
            ],
        ],
        [
            (0x256C, 0x006C),
            [
                WacomStylus.Generic.PEN_3BTN_WITH_ERASER,
                WacomStylus.Generic.ERASER_3BTN,
            ],
        ],
    ],
)
def test_generic_pens(db, usbid, expected):
    # Inspiroy 2 has a generic-pen-no-eraser
    device = db.new_from_usbid(*usbid)
    assert device is not None

    nstyli = ctypes.c_int()
    styli = device.get_supported_styli(ctypes.byref(nstyli))
    s = [WacomStylus.Generic(id) for id in styli[: nstyli.value]]

    assert sorted(s) == sorted(expected)


@pytest.mark.parametrize(
    "bus,vid,pid",
    [
        (WacomBustype.USB, 0x56A, 0xBC),
        (WacomBustype.BLUETOOTH, 0x56A, 0xBD),
        (WacomBustype.UNKNOWN, 0x56A, 0xBD),
    ],
)
def test_new_from_builder_ids(db, bus, vid, pid):
    match = WacomBuilder.create(bus=bus, usbid=(vid, pid))
    device = db.new_from_builder(match)

    assert device is not None
    assert device.vendor_id == vid
    assert device.product_id == pid
    if bus != WacomBustype.UNKNOWN:
        assert device.bustype == bus
    else:
        # unkonwn bustype means "search for it" and
        # for this test that's bluetooth:
        # 0x56a/0bd is a bluetooth tablet
        assert device.bustype == WacomBustype.BLUETOOTH


def test_new_from_builder_empty(db):
    builder = WacomBuilder.create()
    device = db.new_from_builder(builder)
    assert device is None, f"Unexpected device: {device.name}"


def test_new_from_builder_device_name(db):
    builder = WacomBuilder.create(device_name="Wacom Bamboo Pen")
    device = db.new_from_builder(builder)
    assert device is not None

    # Fallback device with name override
    builder.device_name = "does not exist"
    device = db.new_from_builder(builder, fallback=WacomDatabase.Fallback.GENERIC)
    assert device is not None
    assert device.name == "does not exist"


def test_new_from_builder_uniq(db):
    builder = WacomBuilder.create(uniq="OEM02_T18e")
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "GAOMON S620"

    # uniq + match name triggers normal builder
    # but since vid/pid isn't set this does not find a match
    builder = WacomBuilder.create(uniq="OEM02_T18e")
    builder.match_name = "GAOMON Gaomon Tablet Pen"
    device = db.new_from_builder(builder)
    assert device is None

    # Once we set the vid/pid we get a match
    builder.usbid = (0x256C, 0x6D)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "GAOMON S620"

    # uniq + name triggers normal builder but we don't have a code path
    # for that, it's a normal nameless match and
    # since vid/pid isn't set this does not find a match
    builder = WacomBuilder.create(uniq="OEM02_T18e")
    builder.device_name = "GAOMON S620"
    device = db.new_from_builder(builder)
    assert device is None


def test_exact_matches(custom_datadir):
    USBID = (0x1234, 0x5678)
    UNIQ = "uniqval"
    NAME = "nameval"

    # A device match with uniq but no name
    matches = ["usb|1234|5678||uniqval"]
    TabletFile(name="UniqOnly", matches=matches).write_to(
        custom_datadir / "uniq.tablet"
    )

    # A device match with a name but no uniq
    matches = ["usb|1234|5678|nameval"]
    TabletFile(name="NameOnly", matches=matches).write_to(
        custom_datadir / "name.tablet"
    )

    # A device match with both
    matches = ["usb|1234|5678|nameval|uniqval"]
    TabletFile(name="Both", matches=matches).write_to(custom_datadir / "both.tablet")

    db = WacomDatabase(path=custom_datadir)

    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "UniqOnly"

    builder = WacomBuilder.create(usbid=USBID, match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "NameOnly"

    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ, match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "Both"


def test_prefer_uniq_over_name(custom_datadir):
    USBID = (0x1234, 0x5678)
    UNIQ = "uniqval"
    NAME = "nameval"

    # A device match with uniq but no name
    matches = ["usb|1234|5678||uniqval"]
    TabletFile(name="UniqOnly", matches=matches).write_to(
        custom_datadir / "uniq.tablet"
    )

    # A device match with a name but no uniq
    matches = ["usb|1234|5678|nameval"]
    TabletFile(name="NameOnly", matches=matches).write_to(
        custom_datadir / "name.tablet"
    )

    db = WacomDatabase(path=custom_datadir)

    # name and uniq set in our match but we don't have a device with both.
    # Prefer the uniq match over the name match
    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ, match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "UniqOnly"

    # If we have a uniq in our match but none of the DeviceMatches
    # have that, fall back to name only
    builder = WacomBuilder.create(usbid=USBID, uniq="whatever", match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "NameOnly"

    # If we have a name in our match but none of the DeviceMatches
    # have that, fall back to uniq only
    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ, match_name="whatever")
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "UniqOnly"


def test_dont_ignore_exact_matches(custom_datadir):
    USBID = (0x1234, 0x5678)
    UNIQ = "uniqval"
    NAME = "nameval"

    # A device match with both
    matches = ["usb|1234|5678|nameval|uniqval"]
    TabletFile(name="Both", matches=matches).write_to(custom_datadir / "both.tablet")

    db = WacomDatabase(path=custom_datadir)

    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ, match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "Both"

    # Our DeviceMatch has both uniq and name set, so only match
    # when *both* match
    builder = WacomBuilder.create(usbid=USBID, uniq=UNIQ, match_name="whatever")
    device = db.new_from_builder(builder)
    assert device is None

    builder = WacomBuilder.create(usbid=USBID, uniq="whatever", match_name=NAME)
    device = db.new_from_builder(builder)
    assert device is None


# Emulates the behavior of new_from_path for an unknown device but without
# uinput devices
@pytest.mark.parametrize(
    "fallback", (WacomDatabase.Fallback.NONE, WacomDatabase.Fallback.GENERIC)
)
@pytest.mark.parametrize("bustype", (WacomBustype.USB, WacomBustype.BLUETOOTH))
def test_new_unknown_device_with_fallback(custom_datadir, fallback, bustype):
    USBID = (0x1234, 0x5678)
    NAME = "nameval"
    db = WacomDatabase(path=custom_datadir)
    builder = WacomBuilder.create(
        usbid=USBID, bus=bustype, match_name=NAME, device_name=NAME
    )

    device = db.new_from_builder(builder, fallback=fallback)
    if fallback:
        assert device is not None
        match = device.match
        assert match.decode("utf-8") == "generic"
        # Generic device always has 0, 0, 0 triple for bus/vid/pid
        assert device.bustype == WacomBustype.UNKNOWN
        assert device.vendor_id == 0
        assert device.product_id == 0
        assert device.name == NAME
    else:
        assert device is None


def create_uinput(name, vid, pid, bustype=0x3):
    libevdev = pytest.importorskip("libevdev")
    dev = libevdev.Device()
    dev.name = name
    dev.id = {"bustype": bustype, "vendor": vid, "product": pid}
    dev.enable(
        libevdev.EV_ABS.ABS_X,
        libevdev.InputAbsInfo(minimum=0, maximum=10000, resolution=200),
    )
    dev.enable(
        libevdev.EV_ABS.ABS_Y,
        libevdev.InputAbsInfo(minimum=0, maximum=10000, resolution=200),
    )
    dev.enable(libevdev.EV_KEY.BTN_STYLUS)
    dev.enable(libevdev.EV_KEY.BTN_TOOL_PEN)
    try:
        return dev.create_uinput_device()
    except OSError as e:
        pytest.skip(f"Failed to create uinput device: {e}")


@pytest.mark.parametrize(
    "fallback", (WacomDatabase.Fallback.NONE, WacomDatabase.Fallback.GENERIC)
)
def test_new_from_path_known_device(db, fallback):
    name = "Wacom Intuos4 WL"
    vid = 0x056A
    pid = 0x00BC
    uinput = create_uinput(name, vid, pid)

    dev = db.new_from_path(
        uinput.devnode, fallback=fallback
    )  # fallback has no effect here
    assert dev is not None
    assert dev.name == name
    assert dev.vendor_id == vid
    assert dev.product_id == pid


@pytest.mark.parametrize("bustype", WacomBustype)
@pytest.mark.parametrize(
    "fallback", (WacomDatabase.Fallback.NONE, WacomDatabase.Fallback.GENERIC)
)
def test_new_from_path_unknown_device(db, fallback, bustype):
    bus = {
        WacomBustype.UNKNOWN: 0,
        WacomBustype.USB: 0x3,
        WacomBustype.BLUETOOTH: 0x5,
        WacomBustype.I2C: 0x18,
    }[bustype]

    name = "Unknown device"
    vid = 0x1234
    pid = 0xABAC
    uinput = create_uinput(name, vid, pid, bustype=bus)

    dev = db.new_from_path(
        uinput.devnode, fallback=fallback
    )  # fallback has no effect here
    if fallback == WacomDatabase.Fallback.NONE:
        assert dev is None
    else:
        assert dev is not None
        assert dev.name == name
        assert dev.vendor_id == 0
        assert dev.product_id == 0
        assert dev.bustype == 0  # fallback device is always bustype 0


@pytest.mark.parametrize(
    "feature",
    ("Ring", "Strip", "Dial"),
)
@pytest.mark.parametrize("count", (1, 2))
def test_button_modeswitch(custom_datadir, feature, count):
    USBID = (0x1234, 0x5678)

    extra = {
        "Buttons": {
            "Left": "A;B;C;",
            "Right": "D;E;F;",
            feature: "A",
        },
        "Features": {
            f"{feature}s": 4,
        },
    }
    if count > 1:
        extra["Buttons"][f"{feature}2"] = "D"

    TabletFile(
        name="some tablet",
        matches=[f"usb|{USBID[0]:04x}|{USBID[1]:04x}"],
        extra=extra,
    ).write_to(custom_datadir / "led.tablet")

    db = WacomDatabase(path=custom_datadir)
    builder = WacomBuilder.create(usbid=USBID)
    device = db.new_from_builder(builder)
    assert device is not None

    expected_flag = {
        "Ring": WacomDevice.ButtonFlags.RING_MODESWITCH,
        "Strip": WacomDevice.ButtonFlags.TOUCHSTRIP_MODESWITCH,
        "Dial": WacomDevice.ButtonFlags.DIAL_MODESWITCH,
    }[feature]

    flags = device.button_flags("A")
    assert expected_flag in flags

    for b in "BCDEF":
        flags = device.button_flags(b)
        assert expected_flag not in flags

    expected_flag = {
        "Ring": WacomDevice.ButtonFlags.RING2_MODESWITCH,
        "Strip": WacomDevice.ButtonFlags.TOUCHSTRIP2_MODESWITCH,
        "Dial": WacomDevice.ButtonFlags.DIAL2_MODESWITCH,
    }[feature]

    flags = device.button_flags("D")
    if count > 1:
        assert expected_flag in flags
    else:
        assert expected_flag not in flags

    for b in "ABCEF":
        flags = device.button_flags(b)
        assert expected_flag not in flags


@pytest.mark.parametrize(
    "feature",
    ("Ring", "Strip", "Dial"),
)
@pytest.mark.parametrize("count", (1, 2))
def test_status_leds(custom_datadir, feature, count):
    USBID = (0x1234, 0x5678)

    extra = {
        "Buttons": {
            "Left": "A;B;C;",
            "Right": "D;E;F;",
            feature: "A",
        },
        "Features": {
            "StatusLEDs": f"{feature};{feature}2" if count > 1 else f"{feature}",
            f"{feature}s": 4,
        },
    }
    if count > 1:
        extra["Buttons"][f"{feature}2"] = "D"

    TabletFile(
        name="some tablet",
        matches=[f"usb|{USBID[0]:04x}|{USBID[1]:04x}"],
        extra=extra,
    ).write_to(custom_datadir / "led.tablet")

    db = WacomDatabase(path=custom_datadir)
    builder = WacomBuilder.create(usbid=USBID)
    device = db.new_from_builder(builder)
    assert device is not None

    expected = [
        {
            "Ring": WacomStatusLed.RING,
            "Strip": WacomStatusLed.TOUCHSTRIP,
            "Dial": WacomStatusLed.DIAL,
        }[feature]
    ]

    if count > 1:
        expected.append(
            {
                "Ring": WacomStatusLed.RING2,
                "Strip": WacomStatusLed.TOUCHSTRIP2,
                "Dial": WacomStatusLed.DIAL2,
            }[feature]
        )

    leds = device.status_leds
    assert sorted(leds) == sorted(expected)

    led_group = device.button_led_group("A")
    assert led_group == 0

    led_group = device.button_led_group("D")
    if count > 1:
        assert led_group == 1
    else:
        assert led_group == -1

    for b in "BCEF":
        led_group = device.button_led_group(b)
        assert led_group == -1


def test_nonwacom_stylus_ids(tmp_path):
    styli = StylusFile.default()
    s1 = StylusEntry(
        id="0x1234:0xabcd",
        name="ABC Pen",
        group="notwacom",
        paired_stylus_ids=["0x1234:0x9876"],
    )
    s2 = StylusEntry(
        id="0x1234:0x9876",
        name="9876 Pen",
        group="notwacom",
        paired_stylus_ids=["0x1234:0xabcd"],
        eraser_type="Invert",
    )
    styli.entries.append(s1)
    styli.entries.append(s2)
    styli.write_to_dir(tmp_path)

    # matches our nonwacom group
    TabletFile(
        name="ABC Group Tablet",
        matches=["usb|9999|abcd"],
        styli=["@notwacom"],
    ).write_to(tmp_path / "group.tablet")

    # matches one nonwacom styli and the default generic ones
    TabletFile(
        name="ABC Tablet",
        matches=["usb|8888|abcd"],
        styli=[s2.id, "@generic-with-eraser"],
    ).write_to(tmp_path / "abc.tablet")

    db = WacomDatabase(path=tmp_path)

    builder = WacomBuilder.create(usbid=(0x9999, 0xABCD))
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "ABC Group Tablet"
    styli = device.get_styli()
    assert len(styli) == 2
    assert styli[0].vendor_id == 0x1234
    assert styli[1].vendor_id == 0x1234
    # Order of styli is undefined
    assert styli[0].tool_id == 0xABCD or styli[1].tool_id == 0xABCD
    assert styli[0].tool_id == 0x9876 or styli[1].tool_id == 0x9876
    assert sum(s.is_eraser for s in styli) == 1
    for s in filter(lambda s: s.is_eraser, styli):
        assert s.eraser_type == WacomEraserType.INVERT

    assert sum(s.is_eraser is True for s in styli) == 1

    paired0 = styli[0].get_paired_styli()
    paired1 = styli[1].get_paired_styli()
    assert len(paired0) == 1
    assert len(paired1) == 1
    assert paired0[0].vendor_id == styli[1].vendor_id
    assert paired0[0].tool_id != styli[1].vendor_id
    assert paired1[0].vendor_id == styli[0].vendor_id
    assert paired1[0].tool_id != styli[0].vendor_id

    builder = WacomBuilder.create(usbid=(0x8888, 0xABCD))
    device = db.new_from_builder(builder)
    assert device is not None
    assert device.name == "ABC Tablet"
    styli = device.get_styli()
    assert len(styli) == 3  # 1 non-wacom, 2 generic ones
    # Order of styli is undefined
    assert sum(s.vendor_id == 0x1234 and s.tool_id == 0x9876 for s in styli) == 1
    assert sum(s.vendor_id == 0 and s.tool_id == 0xAFFFE for s in styli) == 1
    assert sum(s.vendor_id == 0 and s.tool_id == 0xAFFFF for s in styli) == 1


def test_list_styli_from_database(tmp_path):
    styli = StylusFile.default()
    styli.entries.append(
        StylusEntry(
            id="0x56a:0x802",
            name="Wacom Pen",
        )
    )
    styli.entries.append(
        StylusEntry(
            id="0x1234:0xabcd",
            name="OtherVendor Pen",
        )
    )
    styli.write_to_dir(tmp_path)
    # Two tablets referencing the same styli to verify no duplicates
    TabletFile(
        name="Tablet A",
        matches=["usb|1234|abcd"],
        styli=["0x56a:0x802", "0x1234:0xabcd", "@generic-with-eraser"],
    ).write_to(tmp_path / "tablet-a.tablet")
    TabletFile(
        name="Tablet B",
        matches=["usb|5678|ef01"],
        styli=["0x56a:0x802", "@generic-with-eraser"],
    ).write_to(tmp_path / "tablet-b.tablet")

    db = WacomDatabase(path=tmp_path)
    all_styli = db.list_styli()
    assert len(all_styli) > 0

    # Every stylus should have valid vendor_id and tool_id
    for s in all_styli:
        assert isinstance(s, WacomStylus)

    # Check that specific styli are present
    ids = [(s.vendor_id, s.tool_id) for s in all_styli]
    assert (0x56A, 0x802) in ids
    assert (0x1234, 0xABCD) in ids
    assert (0x0, 0xAFFFF) in ids
    assert (0x0, 0xAFFFE) in ids

    # No duplicates: should contain exactly 4 unique styli even though
    # two tablets reference the same ones
    assert len(ids) == len(set(ids))
    assert len(all_styli) == 4

    # Should be sorted by (vendor_id, tool_id)
    assert ids == sorted(ids)


def test_list_styli_from_database_matches_get_styli(tmp_path):
    """Every stylus from get_styli should be present in list_styli."""
    styli = StylusFile.default()
    styli.entries.append(
        StylusEntry(id="0x56a:0x802", name="Wacom Pen"),
    )
    styli.write_to_dir(tmp_path)
    TabletFile(
        name="Test Tablet",
        matches=["usb|1234|abcd"],
        styli=["0x56a:0x802", "@generic-with-eraser"],
    ).write_to(tmp_path / "test.tablet")

    db = WacomDatabase(path=tmp_path)
    all_styli = db.list_styli()
    all_ids = {(s.vendor_id, s.tool_id) for s in all_styli}

    builder = WacomBuilder.create(usbid=(0x1234, 0xABCD))
    device = db.new_from_builder(builder)
    assert device is not None

    for s in device.get_styli():
        assert (s.vendor_id, s.tool_id) in all_ids


def test_load_xdg_config_home(monkeypatch, tmp_path, custom_datadir):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path.absolute()))

    xdg = tmp_path / "libwacom"
    xdg.mkdir()

    usbid = (0x1234, 0x5678)
    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
    TabletFile(name="XDGTablet", matches=matches).write_to(xdg / "uniq.tablet")

    StylusFile.default().write_to_dir(xdg)

    # This should load from system *and* XDG. system files could
    # interfere with our test or may not exist but unfortunately we can't
    # chroot for the test. it should be good enough this way anyway.
    db = WacomDatabase()
    builder = WacomBuilder.create(usbid=usbid)
    device = db.new_from_builder(builder)
    assert device is not None and device.name == "XDGTablet"


@pytest.mark.parametrize(
    "override,value",
    (
        ("axes", "Tilt;"),
        ("buttons", 3),
        ("has_lens", True),
        ("has_wheel", True),
        ("stylus_type", WacomStylusType.AIRBRUSH),
        ("eraser_type", WacomEraserType.INVERT),
    ),
)
def test_alias_of(custom_datadir, override, value):
    usbid = (0x1234, 0x5678)
    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
    TabletFile(name="XDGTablet", matches=matches, styli=["@happy-aliases"]).write_to(
        custom_datadir / "uniq.tablet"
    )

    stylusfile = StylusFile.default()
    stylusfile.entries.append(
        StylusEntry(
            id="0x1234:0xabcd",
            name="To be aliased",
            group="happy-aliases",
        )
    )

    alias = StylusEntry(
        id="0x1234:0xffff",
        name="AliasTester",
        alias_of="0x1234:0xabcd",
        group="happy-aliases",
    )
    if override == "stylus_type":
        setattr(alias, override, "Airbrush")
    elif override == "eraser_type":
        setattr(alias, override, "Invert")
    else:
        setattr(alias, override, value)

    stylusfile.entries.append(alias)
    stylusfile.write_to_dir(custom_datadir)

    db = WacomDatabase(path=custom_datadir)
    builder = WacomBuilder.create(usbid=usbid)
    device = db.new_from_builder(builder)
    assert device is not None and device.name == "XDGTablet"

    styli = device.get_styli()
    assert len(styli) == 2

    original = styli[0]
    aliasing = styli[1]

    assert original.vendor_id == 0x1234
    assert original.tool_id == 0xABCD
    assert original.num_buttons == 2
    assert original.axes == [
        WacomAxisType.TILT,
        WacomAxisType.DISTANCE,
        WacomAxisType.PRESSURE,
    ]
    assert not original.has_lens
    assert not original.has_wheel
    assert original.eraser_type == WacomEraserType.NONE

    assert aliasing.vendor_id == 0x1234
    assert aliasing.tool_id == 0xFFFF
    if override == "buttons":
        assert aliasing.num_buttons == value
    else:
        assert aliasing.num_buttons == 2

    if override == "axes":
        assert aliasing.axes == [WacomAxisType.TILT]
    else:
        assert aliasing.axes == [
            WacomAxisType.TILT,
            WacomAxisType.DISTANCE,
            WacomAxisType.PRESSURE,
        ]

    if override == "stylus_type":
        assert aliasing.stylus_type == value
    else:
        assert aliasing.stylus_type == WacomStylusType.GENERAL

    if override == "eraser_type":
        assert aliasing.eraser_type == value
    else:
        assert aliasing.eraser_type == WacomEraserType.NONE

    if override == "has_lens":
        assert aliasing.has_lens is value
    else:
        assert aliasing.has_lens is False

    if override == "has_wheel":
        assert aliasing.has_wheel is value
    else:
        assert aliasing.has_wheel is False


@pytest.mark.parametrize("alias", ("0x1234:0xaaaa", "0xabcd", "banana"))
def test_alias_of_invalid(custom_datadir, alias):
    usbid = (0x1234, 0x5678)
    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
    TabletFile(name="XDGTablet", matches=matches, styli=["@happy-aliases"]).write_to(
        custom_datadir / "uniq.tablet"
    )

    stylusfile = StylusFile.default()
    stylusfile.entries = []
    stylusfile.entries.append(
        StylusEntry(
            id="0x1234:0xabcd",
            name="To be aliased",
            group="happy-aliases",
        )
    )

    alias = StylusEntry(
        id="0x1234:0xffff",
        name="AliasTester",
        alias_of=alias,
        group="happy-aliases",
    )
    stylusfile.entries.append(alias)
    stylusfile.write_to_dir(custom_datadir)

    db = WacomDatabase(path=custom_datadir)
    builder = WacomBuilder.create(usbid=usbid)
    device = db.new_from_builder(builder)
    assert device is not None and device.name == "XDGTablet"

    styli = device.get_styli()
    assert len(styli) == 1

    original = styli[0]
    assert original.vendor_id == 0x1234
    assert original.tool_id == 0xABCD
    assert original.num_buttons == 2
    assert original.axes == [
        WacomAxisType.TILT,
        WacomAxisType.DISTANCE,
        WacomAxisType.PRESSURE,
    ]
    assert not original.has_lens
    assert not original.has_wheel
    assert original.eraser_type == WacomEraserType.NONE



#!/usr/bin/env python3
#
# Copyright © 2012 Red Hat, Inc.
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without
# fee, provided that the above copyright notice appear in all copies
# and that both that copyright notice and this permission notice
# appear in supporting documentation, and that the name of Red Hat
# not be used in advertising or publicity pertaining to distribution
# of the software without specific, written prior permission.  Red
# Hat makes no representations about the suitability of this software
# for any purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import argparse
import configparser
import os
import sys
import subprocess
import tempfile
from pathlib import Path


def xdg_dir():
    return Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config")) / "libwacom"


class Tablet(object):
    def __init__(self, name, bus, vid, pid):
        self.name = name
        self.bus = bus
        self.vid = vid  # Note: this is a string
        self.pid = pid  # Note: this is a string
        self.has_touch = False
        self.has_pad = False
        self.is_touchscreen = False

        # We have everything in strings so let's use that for sorting later
        # This will sort bluetooth before usb but meh
        self.cmpstr = ":".join((bus, vid, pid, name))

    def __lt__(self, other):
        return self.cmpstr < other.cmpstr

    def __str__(self):
        return f"{self.bus}:{self.vid}:{self.pid}:{self.name}"


class HWDBFile:
    def __init__(self):
        self.tablets = []

    def _tablet_entry(self, tablet):
        vid = tablet.vid.upper()
        pid = tablet.pid.upper()
        bustypes = {
            "usb": "0003",
            "bluetooth": "0005",
        }
        # serial devices have their own rules, so we skip anything that
        # doesn't have straight conversion
        try:
            bus = bustypes[tablet.bus]
        except KeyError:
            return

        match = f"b{bus}v{vid}p{pid}"
        entries = {"*": ["ID_INPUT=1", "ID_INPUT_TABLET=1", "ID_INPUT_JOYSTICK=0"]}
        if tablet.has_touch:
            if tablet.is_touchscreen:
                entries["* Finger"] = ["ID_INPUT_TOUCHSCREEN=1"]
            else:
                entries["* Finger"] = ["ID_INPUT_TOUCHPAD=1"]

        if tablet.has_pad:
            entries["* Pad"] = ["ID_INPUT_TABLET_PAD=1"]

        # Non-Wacom devices often have a Keyboard node instead of a Pad
        # device. If they share the USB ID with the tablet, we likely just
        # assigned ID_INPUT_TABLET to a keyboard device - and libinput refuses
        # to accept those.
        # Let's add a generic exclusion rule for anything we know of with a
        # Keyboard device name.
        entries["* Keyboard"] = ["ID_INPUT_TABLET=0", "ID_INPUT_TABLET_PAD=0"]
        entries["* Mouse"] = ["ID_INPUT_TABLET=0", "ID_INPUT_TABLET_PAD=0"]

        lines = [f"# {tablet.name}"]
        for name, props in entries.items():
            lines.append(f"libwacom:name:{name}:input:{match}*")
            lines.extend([f" {p}" for p in props])
            lines.append("")

        return "\n".join(lines)

    def print(self, file=sys.stdout):
        header = (
            "# hwdb entries for libwacom supported devices",
            "# This file is generated by libwacom, do not edit",
            "#",
            "# The lookup key is a contract between the udev rules and the hwdb entries.",
            "# It is not considered public API and may change.",
            "",
        )
        print("\n".join(header), file=file)

        for t in self.tablets:
            entry = self._tablet_entry(t)
            if entry:
                print(entry, file=file)


class TabletDatabase:
    def __init__(self, path):
        self.path = path
        self.tablets = sorted(self._load(path))

    def _load(self, path):
        for file in Path(path).glob("*.tablet"):
            config = configparser.ConfigParser()
            config.read(file)
            for match in config["Device"]["DeviceMatch"].split(";"):
                # ignore trailing semicolons
                if not match or match == "generic":
                    continue

                # For hwdb entries we don't care about name matches,
                # it'll just result in duplicate ID_INPUT_TABLET assignments
                # for tablets with re-used usbids and that doesn't matter
                try:
                    bus, vid, pid, *_ = match.split("|")
                except ValueError as e:
                    print(f"Failed to process match {match} in {file}", file=sys.stderr)
                    raise e

                name = config["Device"]["Name"]
                t = Tablet(name, bus, vid, pid)

                try:
                    t.has_touch = config["Features"]["Touch"].lower() == "true"
                    if t.has_touch:
                        integration = config["Device"]["IntegratedIn"]
                        t.is_touchscreen = (
                            "Display" in integration or "System" in integration
                        )
                except KeyError:
                    pass
                t.has_pad = any(config.has_section(s) for s in ["Buttons", "Keys"])
                yield t


# Guess the udev directory based on path. For the case of /usr/share, the
# udev directory is probably in /usr/lib so let's fallback to that.
def find_udev_base_dir(paths):
    for path in paths:
        for parent in path.parents:
            d = Path(parent / "udev" / "rules.d")
            if d.exists():
                return d.parent

        # /usr/share but also any custom prefixes
        for parent in path.parents:
            d = Path(parent / "lib" / "udev" / "rules.d")
            if d.exists():
                return d.parent

    raise FileNotFoundError(paths)


# udev's behaviour is that where a file X exists in two locations,
# only the highest-precedence one is read. Our files are supposed to be
# complimentary to the system-installed ones (which default to
# 65-libwacom.hwdb) so we bump the filename number.
def guess_hwdb_filename(basedir):
    hwdbdir = Path(basedir) / "hwdb.d"
    if not hwdbdir.exists():
        raise FileNotFoundError(hwdbdir)

    fname = hwdbdir / "66-libwacom.hwdb"
    return fname


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Update the system according to the current set of tablet data files"
    )
    parser.add_argument(
        "path",
        nargs="?",
        type=Path,
        default=None,
        help="Directory to load .tablet files from. Default: $XDG_CONFIG_HOME/libwacom and /etc/libwacom",
    )
    # buildsystem-mode is what we use from meson, it changes the
    # the behavior to just generate the file and print it
    parser.add_argument(
        "--buildsystem-mode",
        action="store_true",
        default=False,
        help="be used by the build system only",
    )
    parser.add_argument(
        "--skip-systemd-hwdb-update",
        action="store_true",
        default=False,
        help="Do not run systemd-hwdb --update (Note: updates to tablet files will not be reflected in udev)",
    )
    parser.add_argument(
        "--udev-base-dir",
        type=Path,
        default=None,
        help="The udev base directory (default: guessed based on the path)",
    )
    ns = parser.parse_args()

    if ns.path is None:
        paths = [xdg_dir(), Path("/etc/libwacom")]
    else:
        paths = [ns.path]

    # Reverse the list so the most important one is last and takes precedence
    paths.reverse()
    dbs = [TabletDatabase(p) for p in paths]

    hwdb = HWDBFile()
    # Bamboo and Intuos devices connected to the system via Wacom's
    # Wireless Accessory Kit appear to udev as having the PID of the
    # dongle rather than the actual tablet. Make sure we properly tag
    # such devices.
    #
    # We only really care about this in the official hwdb files
    if ns.buildsystem_mode:
        wwak = Tablet("Wacom Wireless Accessory Kit", "usb", "056A", "0084")
        wwak.has_pad = True
        wwak.has_touch = True
        hwdb.tablets.append(wwak)

    for db in dbs:
        hwdb.tablets.extend(db.tablets)

    if ns.buildsystem_mode:
        hwdb.print()
    else:
        if os.geteuid() == 0:
            print(
                "WARNING: Running this command as root will not pick up .tablet files in $XDG_CONFIG_HOME/libwacom"
            )

        try:
            udevdir = ns.udev_base_dir or find_udev_base_dir(paths)
            hwdbfile = guess_hwdb_filename(udevdir)

            with tempfile.NamedTemporaryFile(
                mode="w+",
                prefix=f"{hwdbfile.name}-",
                suffix=".tmp",
                dir=tempfile.gettempdir(),
                encoding="utf-8",
                delete=True,
            ) as fd:
                hwdb.print(fd)
                fd.flush()
                os.fsync(fd.fileno())
                print(f"Using sudo to copy hwdb file to {hwdbfile}")
                subprocess.run(
                    ["sudo", "cp", "--", fd.name, str(hwdbfile.absolute())],
                    check=True,
                )

            if not ns.skip_systemd_hwdb_update:
                print("Using sudo to run systemd-hwdb update")
                subprocess.run(
                    ["sudo", "systemd-hwdb", "update"],
                    capture_output=True,
                    check=True,
                    text=True,
                )
            print("Finished, please unplug and replug your device")
        except PermissionError as e:
            print(f"{e}, please run me as root")
        except FileNotFoundError as e:
            print(f"Unable to find udev base directory: {e}")
        except subprocess.CalledProcessError as e:
            print(f"hwdb update failed: {e.stderr}")
        except KeyboardInterrupt:
            pass



#!/usr/bin/env python3
#
# This file is formatted with ruff format
#
# Run with pytest:
#    $ export LD_LIBRARY_PATH=$PWD/builddir/
#    $ export MESON_SOURCE_ROOT=$PWD   # optional, defaults to $PWD.
#    $ pytest -v --log-level=DEBUG
#
# Introduction
# ============
#
# This is a Python-based test suite making use of ctypes to test the
# libwacom.so C library.
#
# The main components are:
# - LibWacom: the Python class wrapping libwacom.so via ctypes.
#   This is a manually maintained mapping, any API additions/changes must
#   updated here.
# - WacomDevice, WacomDatabase, ...: pythonic wrappers around the
#   underlying C object.

from ctypes import c_char_p, c_char, c_int, c_uint32, c_void_p
from typing import Optional, Tuple, Type, List
from dataclasses import dataclass
from pathlib import Path

import ctypes
import enum
import itertools
import logging


logger = logging.getLogger(__name__)


PREFIX = "libwacom_"


@dataclass
class _Api:
    name: str
    args: Tuple[Type[ctypes._SimpleCData], ...]
    return_type: Optional[Type[ctypes._SimpleCData]]

    @property
    def basename(self) -> str:
        return self.name.removeprefix(PREFIX)


@dataclass
class _Enum:
    name: str
    value: int

    @property
    def basename(self) -> str:
        return self.name.removeprefix("WACOM_").removeprefix("W")


class GlibC:
    _lib = None

    _api_prototypes: List[_Api] = [
        _Api(name="free", args=(c_void_p,), return_type=None),
    ]

    @staticmethod
    def _cdll():
        # BSD has 7, Linux has 6, Linux/alpha has 6.1
        for libc in ("libc.so.6", "libc.so.7", "libc.so.6.1"):
            try:
                return ctypes.CDLL(libc, use_errno=True)
            except OSError:
                pass
        raise NotImplementedError("Not implemented for other libc.so")

    @classmethod
    def _load(cls):
        cls._lib = cls._cdll()
        for api in cls._api_prototypes:
            func = getattr(cls._lib, api.name)
            func.argtypes = api.args
            func.restype = api.return_type
            setattr(cls, api.basename, func)

    @classmethod
    def instance(cls):
        if cls._lib is None:
            cls._load()
        return cls


class LibWacom:
    """
    libwacom.so wrapper. This is a singleton ctypes wrapper into libwacom.so with
    minimal processing. Example:

    >>> lib = LibWacom.instance()
    >>> ctx = lib.database_new(None)
    >>> lib.wacom_unref(ctx)

    In most cases you probably want to use the ``WacomDevice`` class instead.
    """

    _lib = None

    @staticmethod
    def _cdll():
        return ctypes.CDLL("libwacom.so.9", use_errno=True)

    @classmethod
    def _load(cls):
        cls._lib = cls._cdll()
        for api in cls._api_prototypes:
            func = getattr(cls._lib, api.name)
            func.argtypes = api.args
            func.restype = api.return_type
            setattr(cls, api.basename, func)

        for e in cls._enums:
            setattr(cls, e.basename, e.value)

    @classmethod
    def instance(cls):
        if cls._lib is None:
            cls._load()
        return cls

    _api_prototypes: List[_Api] = [
        _Api(name="libwacom_error_new", args=(c_void_p,), return_type=c_void_p),
        _Api(name="libwacom_error_free", args=(c_void_p,), return_type=None),
        _Api(name="libwacom_error_get_code", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_error_get_message", args=(c_void_p,), return_type=c_char_p),
        _Api(name="libwacom_database_new", args=(), return_type=c_void_p),
        _Api(
            name="libwacom_database_new_for_path",
            args=(c_char_p,),
            return_type=c_void_p,
        ),
        _Api(name="libwacom_database_destroy", args=(c_void_p,), return_type=None),
        _Api(
            name="libwacom_new_from_builder",
            args=(c_void_p, c_void_p, c_int, c_void_p),
            return_type=c_void_p,
        ),
        _Api(
            name="libwacom_new_from_path",
            args=(c_void_p, c_char_p, c_int, c_void_p),
            return_type=c_void_p,
        ),
        _Api(
            name="libwacom_new_from_usbid",
            args=(c_void_p, c_int, c_int, c_void_p),
            return_type=c_void_p,
        ),
        _Api(
            name="libwacom_new_from_name",
            args=(c_void_p, c_char_p, c_void_p),
            return_type=c_void_p,
        ),
        _Api(
            name="libwacom_list_devices_from_database",
            args=(c_void_p, c_void_p),
            return_type=ctypes.POINTER(c_void_p),
        ),
        _Api(
            name="libwacom_list_styli_from_database",
            args=(c_void_p, c_void_p),
            return_type=ctypes.POINTER(c_void_p),
        ),
        _Api(
            name="libwacom_print_device_description",
            args=(c_int, c_void_p),
            return_type=None,
        ),
        _Api(name="libwacom_destroy", args=(c_void_p,), return_type=None),
        _Api(
            name="libwacom_compare", args=(c_void_p, c_void_p, c_int), return_type=c_int
        ),
        _Api(name="libwacom_get_class", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_name", args=(c_void_p,), return_type=c_char_p),
        _Api(name="libwacom_get_model_name", args=(c_void_p,), return_type=c_char_p),
        _Api(
            name="libwacom_get_layout_filename", args=(c_void_p,), return_type=c_char_p
        ),
        _Api(name="libwacom_get_vendor_id", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_match", args=(c_void_p,), return_type=c_char_p),
        _Api(
            name="libwacom_get_matches",
            args=(c_void_p,),
            return_type=ctypes.POINTER(c_void_p),
        ),
        _Api(name="libwacom_get_paired_device", args=(c_void_p,), return_type=c_void_p),
        _Api(name="libwacom_get_product_id", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_width", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_height", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_width_mm", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_height_mm", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_has_stylus", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_has_touch", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_num_buttons", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_num_keys", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_get_supported_styli",
            args=(c_void_p, ctypes.POINTER(c_int)),
            return_type=ctypes.POINTER(c_int),
        ),
        _Api(
            name="libwacom_get_styli",
            args=(c_void_p, c_void_p),
            return_type=ctypes.POINTER(c_void_p),
        ),
        _Api(name="libwacom_has_ring", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_has_ring2", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_num_rings", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_has_touchswitch", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_ring_num_modes", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_ring2_num_modes", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_num_strips", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_strips_num_modes", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_num_dials", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_dial_num_modes", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_get_dial2_num_modes", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_get_status_leds",
            args=(c_void_p, c_void_p),
            return_type=ctypes.POINTER(ctypes.c_int),
        ),
        _Api(
            name="libwacom_get_button_led_group",
            args=(c_void_p, c_char),
            return_type=c_int,
        ),
        _Api(name="libwacom_is_builtin", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_is_reversible", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_get_integration_flags", args=(c_void_p,), return_type=c_int
        ),
        _Api(name="libwacom_get_bustype", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_get_button_flag", args=(c_void_p, c_char), return_type=c_int
        ),
        _Api(
            name="libwacom_get_button_evdev_code",
            args=(c_void_p, c_char),
            return_type=c_int,
        ),
        _Api(
            name="libwacom_get_button_modeswitch_mode",
            args=(c_void_p, c_char),
            return_type=c_int,
        ),
        _Api(
            name="libwacom_stylus_get_for_id",
            args=(c_void_p, c_int),
            return_type=c_void_p,
        ),
        _Api(name="libwacom_stylus_get_id", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_get_vendor_id", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_get_name", args=(c_void_p,), return_type=c_char_p),
        _Api(
            name="libwacom_stylus_get_paired_ids",
            args=(c_void_p, c_void_p),
            return_type=c_void_p,
        ),
        _Api(
            name="libwacom_stylus_get_num_buttons", args=(c_void_p,), return_type=c_int
        ),
        _Api(name="libwacom_stylus_has_eraser", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_is_eraser", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_has_lens", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_has_wheel", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_get_axes", args=(c_void_p,), return_type=c_int),
        _Api(name="libwacom_stylus_get_type", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_stylus_get_eraser_type", args=(c_void_p,), return_type=c_int
        ),
        _Api(
            name="libwacom_stylus_get_paired_styli",
            args=(c_void_p, c_void_p),
            return_type=ctypes.POINTER(c_void_p),
        ),
        _Api(
            name="libwacom_print_stylus_description",
            args=(c_int, c_void_p),
            return_type=None,
        ),
        _Api(name="libwacom_builder_new", args=(), return_type=c_void_p),
        _Api(name="libwacom_builder_destroy", args=(c_void_p,), return_type=None),
        _Api(
            name="libwacom_builder_set_device_name",
            args=(c_void_p, c_char_p),
            return_type=None,
        ),
        _Api(
            name="libwacom_builder_set_match_name",
            args=(c_void_p, c_char_p),
            return_type=None,
        ),
        _Api(
            name="libwacom_builder_set_uniq",
            args=(c_void_p, c_char_p),
            return_type=None,
        ),
        _Api(
            name="libwacom_builder_set_bustype",
            args=(c_void_p, c_int),
            return_type=None,
        ),
        _Api(
            name="libwacom_builder_set_usbid",
            args=(c_void_p, c_int, c_int),
            return_type=None,
        ),
        _Api(name="libwacom_match_get_name", args=(c_void_p,), return_type=c_char_p),
        _Api(name="libwacom_match_get_uniq", args=(c_void_p,), return_type=c_char_p),
        _Api(name="libwacom_match_get_bustype", args=(c_void_p,), return_type=c_int),
        _Api(
            name="libwacom_match_get_product_id", args=(c_void_p,), return_type=c_uint32
        ),
        _Api(
            name="libwacom_match_get_vendor_id", args=(c_void_p,), return_type=c_uint32
        ),
        _Api(
            name="libwacom_match_get_match_string",
            args=(c_void_p,),
            return_type=c_char_p,
        ),
    ]

    _enums: List[_Enum] = [
        _Enum(name="WERROR_NONE", value=0),
        _Enum(name="WERROR_BAD_ALLOC", value=1),
        _Enum(name="WERROR_INVALID_PATH", value=2),
        _Enum(name="WERROR_INVALID_DB", value=3),
        _Enum(name="WERROR_BAD_ACCESS", value=4),
        _Enum(name="WERROR_UNKNOWN_MODEL", value=5),
        _Enum(name="WERROR_BUG_CALLER", value=6),
        _Enum(name="WBUSTYPE_UNKNOWN", value=0),
        _Enum(name="WBUSTYPE_USB", value=1),
        _Enum(name="WBUSTYPE_SERIAL", value=2),
        _Enum(name="WBUSTYPE_BLUETOOTH", value=3),
        _Enum(name="WBUSTYPE_I2C", value=4),
        _Enum(name="WACOM_DEVICE_INTEGRATED_NONE", value=0),
        _Enum(name="WACOM_DEVICE_INTEGRATED_DISPLAY", value=1),
        _Enum(name="WACOM_DEVICE_INTEGRATED_SYSTEM", value=2),
        _Enum(name="WCLASS_UNKNOWN", value=0),
        _Enum(name="WCLASS_INTUOS3", value=1),
        _Enum(name="WCLASS_INTUOS4", value=2),
        _Enum(name="WCLASS_INTUOS5", value=3),
        _Enum(name="WCLASS_CINTIQ", value=4),
        _Enum(name="WCLASS_BAMBOO", value=5),
        _Enum(name="WCLASS_GRAPHIRE", value=6),
        _Enum(name="WCLASS_ISDV4", value=7),
        _Enum(name="WCLASS_INTUOS", value=8),
        _Enum(name="WCLASS_INTUOS2", value=9),
        _Enum(name="WCLASS_PEN_DISPLAYS", value=10),
        _Enum(name="WCLASS_REMOTE", value=2),
        _Enum(name="WSTYLUS_UNKNOWN", value=0),
        _Enum(name="WSTYLUS_GENERAL", value=1),
        _Enum(name="WSTYLUS_INKING", value=2),
        _Enum(name="WSTYLUS_AIRBRUSH", value=3),
        _Enum(name="WSTYLUS_CLASSIC", value=4),
        _Enum(name="WSTYLUS_MARKER", value=5),
        _Enum(name="WSTYLUS_STROKE", value=6),
        _Enum(name="WSTYLUS_PUCK", value=7),
        _Enum(name="WSTYLUS_3D", value=8),
        _Enum(name="WSTYLUS_MOBILE", value=9),
        _Enum(name="WACOM_ERASER_UNKNOWN", value=0),
        _Enum(name="WACOM_ERASER_NONE", value=1),
        _Enum(name="WACOM_ERASER_INVERT", value=2),
        _Enum(name="WACOM_ERASER_BUTTON", value=3),
        _Enum(name="WACOM_BUTTON_NONE", value=0),
        _Enum(name="WACOM_BUTTON_POSITION_LEFT", value=1 << 1),
        _Enum(name="WACOM_BUTTON_POSITION_RIGHT", value=1 << 2),
        _Enum(name="WACOM_BUTTON_POSITION_TOP", value=1 << 3),
        _Enum(name="WACOM_BUTTON_POSITION_BOTTOM", value=1 << 4),
        _Enum(name="WACOM_BUTTON_RING_MODESWITCH", value=1 << 5),
        _Enum(name="WACOM_BUTTON_RING2_MODESWITCH", value=1 << 6),
        _Enum(name="WACOM_BUTTON_TOUCHSTRIP_MODESWITCH", value=1 << 7),
        _Enum(name="WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH", value=1 << 8),
        _Enum(name="WACOM_BUTTON_OLED", value=1 << 9),
        _Enum(name="WACOM_BUTTON_DIAL_MODESWITCH", value=1 << 10),
        _Enum(name="WACOM_BUTTON_DIAL2_MODESWITCH", value=1 << 11),
        _Enum(
            name="WACOM_BUTTON_MODESWITCH",
            value=1 << 5 | 1 << 6 | 1 << 7 | 1 << 8 | 1 << 10 | 1 << 11,
        ),
        _Enum(name="WACOM_BUTTON_DIRECTION", value=1 << 1 | 1 << 2 | 1 << 3 | 1 << 4),
        _Enum(name="WACOM_BUTTON_RINGS_MODESWITCH", value=1 << 5 | 1 << 6),
        _Enum(name="WACOM_BUTTON_TOUCHSTRIPS_MODESWITCH", value=1 << 7 | 1 << 8),
        _Enum(name="WACOM_BUTTON_DIALS_MODESWITCH", value=1 << 10 | 1 << 11),
        _Enum(name="WACOM_AXIS_TYPE_NONE", value=0),
        _Enum(name="WACOM_AXIS_TYPE_TILT", value=1 << 1),
        _Enum(name="WACOM_AXIS_TYPE_ROTATION_Z", value=1 << 2),
        _Enum(name="WACOM_AXIS_TYPE_DISTANCE", value=1 << 3),
        _Enum(name="WACOM_AXIS_TYPE_PRESSURE", value=1 << 4),
        _Enum(name="WACOM_AXIS_TYPE_SLIDER", value=1 << 5),
        _Enum(name="WFALLBACK_NONE", value=0),
        _Enum(name="WFALLBACK_GENERIC", value=1),
        _Enum(name="WCOMPARE_NORMAL", value=0),
        _Enum(name="WCOMPARE_MATCHES", value=1),
        _Enum(name="WACOM_STATUS_LED_UNAVAILABLE", value=0),
        _Enum(name="WACOM_STATUS_LED_RING", value=1),
        _Enum(name="WACOM_STATUS_LED_RING2", value=2),
        _Enum(name="WACOM_STATUS_LED_TOUCHSTRIP", value=3),
        _Enum(name="WACOM_STATUS_LED_TOUCHSTRIP2", value=4),
        _Enum(name="WACOM_STATUS_LED_DIAL", value=1),
        _Enum(name="WACOM_STATUS_LED_DIAL2", value=2),
        _Enum(name="WACOM_MODE_SWITCH_NEXT", value=-1),
        _Enum(name="WACOM_MODE_SWITCH_0", value=0),
        _Enum(name="WACOM_MODE_SWITCH_1", value=1),
        _Enum(name="WACOM_MODE_SWITCH_2", value=2),
        _Enum(name="WACOM_MODE_SWITCH_3", value=3),
    ]


class WacomBustype(enum.IntEnum):
    UNKNOWN = 0x0
    USB = 0x1
    I2C = 0x2
    BLUETOOTH = 0x3


class WacomMatch:
    def __init__(self, match):
        self.match = match
        lib = LibWacom.instance()

        def wrapper(func):
            return lambda *args, **kwargs: func(self.match, *args, **kwargs)

        # Map all device-specifice accessors into respective functions
        for api in lib._api_prototypes:
            allowlist = ["match"]
            if any(api.basename.startswith(n) for n in allowlist):
                func = getattr(lib, api.basename)
                setattr(self, api.basename.removeprefix("match_"), wrapper(func))

    @property
    def name(self) -> Optional[str]:
        name = self.get_name()
        return name.decode("utf-8") if name else None

    @property
    def uniq(self) -> Optional[str]:
        uniq = self.get_uniq()
        return uniq.decode("utf-8") if uniq else None

    @property
    def bustype(self) -> WacomBustype:
        return WacomBustype(self.get_bustype())

    @property
    def vendor_id(self) -> int:
        return self.get_vendor_id()

    @property
    def product_id(self) -> int:
        return self.get_product_id()


class WacomBuilder:
    def __init__(self, builder):
        self.builder = builder
        lib = LibWacom.instance()
        self._device_name = None
        self._match_name = None
        self._uniq = None
        self._usbid = None
        self._bustype = None

        def wrapper(func):
            return lambda *args, **kwargs: func(self.builder, *args, **kwargs)

        # Map all device-specifice accessors into respective functions
        for api in lib._api_prototypes:
            allowlist = ["builder"]
            if any(api.basename.startswith(n) for n in allowlist):
                func = getattr(lib, api.basename)
                setattr(self, api.basename.removeprefix("builder_"), wrapper(func))

    @property
    def device_name(self) -> Optional[str]:
        return self._device_name

    @property
    def match_name(self) -> Optional[str]:
        return self._match_name

    @property
    def uniq(self) -> Optional[str]:
        return self._uniq

    @property
    def bustype(self) -> Optional[WacomBustype]:
        return self._bustype

    @property
    def usbid(self) -> Optional[Tuple[int, int]]:
        return self._usbid

    @bustype.setter
    def bustype(self, bus: WacomBustype):
        self._bustype = bus
        self.set_bustype(bus.value)

    @usbid.setter
    def usbid(self, usbid: Tuple[int, int]):
        self._usbid = usbid
        self.set_usbid(usbid[0], usbid[1])

    @device_name.setter
    def device_name(self, name: str):
        self._device_name = name
        self.set_device_name(name.encode("utf-8"))

    @match_name.setter
    def match_name(self, name: str):
        self._match_name = name
        self.set_match_name(name.encode("utf-8"))

    @uniq.setter
    def uniq(self, uniq: str):
        self._uniq = uniq
        self.set_uniq(uniq.encode("utf-8"))

    @classmethod
    def create(
        cls,
        device_name: Optional[str] = None,
        match_name: Optional[str] = None,
        uniq: Optional[str] = None,
        usbid: Optional[Tuple[int, int]] = None,
        bus: Optional[WacomBustype] = None,
    ) -> "WacomBuilder":
        lib = LibWacom.instance()
        builder = WacomBuilder(lib.builder_new())
        if device_name is not None:
            builder.device_name = device_name
        if match_name is not None:
            builder.match_name = match_name
        if uniq is not None:
            builder.uniq = uniq
        if bus is not None:
            builder.bustype = bus
        if usbid is not None:
            builder.usbid = usbid
        return builder

    def __del__(self):
        lib = LibWacom.instance()
        lib.builder_destroy(self.builder)


class WacomStylusType(enum.IntEnum):
    UNKNOWN = 0
    GENERAL = 1
    INKING = 2
    AIRBRUSH = 3
    CLASSIC = 4
    MARKER = 5
    STROKE = 6
    PUCK = 7
    THREED = 8
    MOBILE = 9


class WacomEraserType(enum.IntEnum):
    UNKNOWN = 0
    NONE = 1
    INVERT = 2
    BUTTON = 3


class WacomAxisType(enum.IntEnum):
    NONE = 0
    TILT = (1 << 1,)
    ROTATION_Z = (1 << 2,)
    DISTANCE = (1 << 3,)
    PRESSURE = (1 << 4,)
    SLIDER = 1 << 5


class WacomStylus:
    class Generic(enum.IntEnum):
        PEN_WITH_ERASER = 0xFFFFF
        ERASER = 0xFFFFE
        PEN_NO_ERASER = 0xFFFFD
        ERASER_3BTN = 0xFFFFC
        PEN_3BTN_WITH_ERASER = 0xFFFFB

    def __init__(self, stylus):
        self.stylus = stylus
        lib = LibWacom.instance()

        def wrapper(func):
            return lambda *args, **kwargs: func(self.stylus, *args, **kwargs)

        # Map all device-specifice accessors into respective functions
        for api in lib._api_prototypes:
            allowlist = ["stylus"]
            if any(api.basename.startswith(n) for n in allowlist):
                denylist = [
                    "stylus_get_paired_styli",
                    "stylus_is_eraser",
                    "stylus_has_lens",
                    "stylus_has_wheel",
                ]
                if all(not api.basename.startswith(n) for n in denylist):
                    func = getattr(lib, api.basename)
                    setattr(self, api.basename.removeprefix("stylus_"), wrapper(func))

    @property
    def name(self):
        return self.get_name().decode("utf-8")

    @property
    def group(self):
        return self.get_group().decode("utf-8")

    @property
    def tool_id(self) -> int:
        return self.get_id()

    @property
    def vendor_id(self) -> int:
        return self.get_vendor_id()

    @property
    def axes(self) -> list[WacomAxisType]:
        axes = self.get_axes()
        return [t for t in WacomAxisType if axes & t]

    @property
    def num_buttons(self) -> int:
        return self.get_num_buttons()

    @property
    def has_lens(self) -> bool:
        lib = LibWacom.instance()
        return lib.stylus_has_lens(self.stylus) != 0

    @property
    def has_wheel(self) -> bool:
        lib = LibWacom.instance()
        return lib.stylus_has_wheel(self.stylus) != 0

    @property
    def is_eraser(self) -> bool:
        lib = LibWacom.instance()
        return lib.stylus_is_eraser(self.stylus) != 0

    @property
    def stylus_type(self) -> WacomStylusType:
        return WacomStylusType(self.get_type())

    @property
    def eraser_type(self) -> WacomEraserType:
        return WacomEraserType(self.get_eraser_type())

    def get_paired_styli(self) -> List["WacomStylus"]:
        lib = LibWacom.instance()
        paired = lib.stylus_get_paired_styli(self.stylus, None)
        styli = [
            WacomStylus(p)
            for p in itertools.takewhile(lambda ptr: ptr is not None, paired)
        ]
        GlibC.instance().free(paired)
        return styli


class WacomStatusLed(enum.IntEnum):
    UNAVAILABLE = -1
    RING = 0
    RING2 = 1
    TOUCHSTRIP = 2
    TOUCHSTRIP2 = 3
    DIAL = 4
    DIAL2 = 5


class WacomDevice:
    """
    Convenience wrapper to make using libwacom a bit more pythonic.
    """

    class IntegrationFlags(enum.IntEnum):
        DISPLAY = 1 << 0
        SYSTEM = 1 << 1

    class ButtonFlags(enum.IntEnum):
        NONE = 0
        POSITION_LEFT = 1 << 1
        POSITION_RIGHT = 1 << 2
        POSITION_TOP = 1 << 3
        POSITION_BOTTOM = 1 << 4
        RING_MODESWITCH = 1 << 5
        RING2_MODESWITCH = 1 << 6
        TOUCHSTRIP_MODESWITCH = 1 << 7
        TOUCHSTRIP2_MODESWITCH = 1 << 8
        OLED = 1 << 9
        DIAL_MODESWITCH = 1 << 10
        DIAL2_MODESWITCH = 1 << 11

        @staticmethod
        def modeswitch_flags() -> List["WacomDevice.ButtonFlags"]:
            return [
                WacomDevice.ButtonFlags.RING_MODESWITCH,
                WacomDevice.ButtonFlags.RING2_MODESWITCH,
                WacomDevice.ButtonFlags.TOUCHSTRIP_MODESWITCH,
                WacomDevice.ButtonFlags.TOUCHSTRIP2_MODESWITCH,
                WacomDevice.ButtonFlags.DIAL_MODESWITCH,
                WacomDevice.ButtonFlags.DIAL2_MODESWITCH,
            ]

    class ModeSwitch(enum.IntEnum):
        NEXT = -1
        MODE_0 = 0
        MODE_1 = 1
        MODE_2 = 2
        MODE_3 = 3

    def __init__(self, device, destroy=True):
        self.device = device
        self._destroy_on_del = destroy

        lib = LibWacom.instance()

        def wrapper(func):
            return lambda *args, **kwargs: func(self.device, *args, **kwargs)

        # Map all device-specifice accessors into respective functions
        for api in lib._api_prototypes:
            allowlist = ["get_", "is_", "has_"]
            if any(api.basename.startswith(n) for n in allowlist):
                denylist = ["get_paired_device", "get_matches", "get_styli"]
                if all(not api.basename.startswith(n) for n in denylist):
                    func = getattr(lib, api.basename)
                    setattr(self, api.basename, wrapper(func))

        # This mashes all enums into the same namespace but oh well
        for e in lib._enums:
            val = getattr(lib, e.basename)
            setattr(self, e.basename, val)

    def get_paired_device(self) -> Optional[WacomMatch]:
        lib = LibWacom.instance()
        match = lib.get_paired_device(self.device)
        return WacomMatch(match) if match else None

    def get_matches(self) -> List[WacomMatch]:
        lib = LibWacom.instance()
        matches = lib.get_matches(self.device)

        return [
            WacomMatch(m)
            for m in itertools.takewhile(lambda ptr: ptr is not None, matches)
        ]

    def get_styli(self) -> List[WacomStylus]:
        lib = LibWacom.instance()
        styli = lib.get_styli(self.device, None)

        return [
            WacomStylus(m)
            for m in itertools.takewhile(lambda ptr: ptr is not None, styli)
        ]

    def __del__(self):
        if self._destroy_on_del:
            lib = LibWacom.instance()
            lib.destroy(self.device)

    @property
    def paired_device(self) -> Optional[WacomMatch]:
        return self.get_paired_device()

    @property
    def name(self):
        return self.get_name().decode("utf-8")

    @property
    def model_name(self):
        model = self.get_model_name()
        return model.decode("utf-8") if model else None

    @property
    def layout_filename(self):
        f = self.get_layout_filename()
        return f.decode("utf-8") if f else None

    @property
    def bustype(self):
        return self.get_bustype()

    @property
    def vendor_id(self):
        return self.get_vendor_id()

    @property
    def product_id(self):
        return self.get_product_id()

    @property
    def width(self):
        return self.get_width()

    @property
    def height(self):
        return self.get_height()

    @property
    def width_mm(self):
        return self.get_width_mm()

    @property
    def height_mm(self):
        return self.get_height_mm()

    @property
    def num_buttons(self):
        return self.get_num_buttons()

    @property
    def num_keys(self):
        return self.get_num_keys()

    @property
    def num_rings(self):
        return self.get_num_rings()

    @property
    def num_strips(self):
        return self.get_num_strips()

    @property
    def num_dials(self):
        return self.get_num_dials()

    @property
    def ring_num_modes(self):
        return self.get_ring_num_modes()

    @property
    def ring2_num_modes(self):
        return self.get_ring2_num_modes()

    @property
    def strip_num_modes(self):
        return self.get_strip_num_modes()

    @property
    def dial_num_modes(self):
        return self.get_dial_num_modes()

    @property
    def match(self):
        return self.get_match()

    @property
    def matches(self):
        return self.get_matches()

    @property
    def integration_flags(self) -> List[IntegrationFlags]:
        flags = self.get_integration_flags()
        return [f for f in WacomDevice.IntegrationFlags if f & flags != 0]

    def button_flags(self, button: str) -> List[ButtonFlags]:
        flags = self.get_button_flag(button.encode("utf-8"))
        return [f for f in WacomDevice.ButtonFlags if f & flags != 0]

    def button_evdev_code(self, button: str) -> int:
        return self.get_button_evdev_code(button.encode("utf-8"))

    def button_modeswitch_mode(self, button: str) -> ModeSwitch:
        mode = self.get_button_modeswitch_mode(button.encode("utf-8"))
        return WacomDevice.ModeSwitch(mode)

    def button_led_group(self, button: str) -> List[ButtonFlags]:
        return self.get_button_led_group(button.encode("utf-8"))

    @property
    def status_leds(self) -> List["WacomStatusLed"]:
        nleds = c_int()
        leds = self.get_status_leds(ctypes.byref(nleds))

        return [WacomStatusLed(l) for l in leds[: nleds.value]]


class WacomDatabase:
    """
    Convenience wrapper to make using libwacom a bit more pythonic.
    """

    class Fallback(enum.IntEnum):
        NONE = 0x0
        GENERIC = 0x1

    def __init__(self, path: Optional[Path] = None):
        lib = LibWacom.instance()
        if path is None:
            self.db = lib.database_new()  # type: ignore
        else:
            self.db = lib.database_new_for_path(str(path).encode("utf-8"))  # type: ignore
        assert self.db is not None

        def wrapper(func):
            return lambda *args, **kwargs: func(self.db, *args, **kwargs)

        for api in lib._api_prototypes:
            prefixes = [
                "new_from_",
                "list_devices_from_database",
                "list_styli_from_database",
            ]
            if any(api.basename.startswith(prefix) for prefix in prefixes):
                func = getattr(lib, api.basename)
                setattr(self, api.name, wrapper(func))

    def __del__(self):
        db = getattr(self, "db", None)
        if db is not None:
            lib = LibWacom.instance()
            lib.database_destroy(db)

    def new_from_name(self, name: str) -> Optional[WacomDevice]:
        device = self.libwacom_new_from_name(name.encode("utf-8"), 0)
        return WacomDevice(device) if device else None

    def new_from_path(
        self, path: str, fallback: Fallback = Fallback.NONE
    ) -> Optional[WacomDevice]:
        device = self.libwacom_new_from_path(path.encode("utf-8"), fallback, 0)
        return WacomDevice(device) if device else None

    def new_from_usbid(self, vid: int, pid: int) -> Optional[WacomDevice]:
        device = self.libwacom_new_from_usbid(vid, pid, 0)
        return WacomDevice(device) if device else None

    def new_from_builder(
        self, builder: WacomBuilder, fallback: Fallback = Fallback.NONE
    ) -> Optional[WacomDevice]:
        device = self.libwacom_new_from_builder(builder.builder, fallback.value, 0)
        return WacomDevice(device) if device else None

    def list_devices(self) -> List[WacomDevice]:
        devices = self.libwacom_list_devices_from_database(self.db, 0)
        devs = [
            WacomDevice(d, destroy=False)
            for d in itertools.takewhile(lambda ptr: ptr is not None, devices)
        ]
        GlibC.instance().free(devices)
        return devs

    def list_styli(self) -> List[WacomStylus]:
        styli = self.libwacom_list_styli_from_database(self.db, 0)
        result = [
            WacomStylus(s)
            for s in itertools.takewhile(lambda ptr: ptr is not None, styli)
        ]
        GlibC.instance().free(styli)
        return result



#!/usr/bin/env python3
#
# This test will only work where /dev/uinput is available.
# This test will reload the hwdb and udev rules
#
# Execute via pytest, it will:
# - load all data files and extract the matches
# - create a uinput device for each match
# - check if that device has the udev properties set we expect

import configparser
import os
from pathlib import Path
import pytest
import logging
import sys
import subprocess
import shutil


@pytest.fixture(scope="session", autouse=True)
def systemd_reload():
    """Make sure our hwdb and udev rules are up-to-date"""

    try:
        hwdb = os.environ.get("LIBWACOM_HWDB_FILE")
        target = Path("/etc/udev/hwdb.d/99-libwacom-pytest.hwdb")
        target.parent.mkdir(exist_ok=True, parents=True)
        if hwdb:
            shutil.copyfile(hwdb, target)
        else:
            import warnings

            warnings.warn("LIBWACOM_HWDB_FILE is not set, using already installed hwdb")

        subprocess.run(["systemd-hwdb", "update"], check=True)

        yield

        if hwdb:
            os.unlink(target)

        subprocess.run(["systemd-hwdb", "update"], check=True)

    except (IOError, FileNotFoundError, subprocess.CalledProcessError) as e:
        # If any of the commands above are not found (most likely the system
        # simply does not use systemd), just skip.
        logging.critical(f"{e}")
        pytest.skip(f"Skipping test: {e}")
    except Exception as e:
        logging.critical(f"{e}")
        pytest.fail(f"Aborting test: {e}")


def pytest_generate_tests(metafunc):
    # for any function that takes a "tablet" argument return a Tablet object
    # filled with exactly one DeviceMatch from the list of all .tablet files
    # in the data dir. Where the tablet also has touch/buttons generate an
    # extra Finger or Pad device
    if "tablet" in metafunc.fixturenames:
        datadir = Path(os.getenv("MESON_SOURCE_ROOT") or ".") / "data"
        tablets = []
        for f in datadir.glob("*.tablet"):
            config = configparser.ConfigParser()
            config.read(f)
            name = config["Device"]["Name"]
            want_pad = config["Device"].get("Buttons", 0)
            want_finger = config["Features"].get("Touch") == "true"
            integrated_in = config["Device"].get("IntegratedIn", "").split(";")
            is_touchscreen = set(integrated_in) & set(["Display", "System"])

            for match in config["Device"]["DeviceMatch"].split(";"):
                if not match or match == "generic":
                    continue

                bus, vid, pid = match.split("|")[:3]  # skip the name part of the match
                if bus not in ["usb", "bluetooth"]:
                    continue

                try:
                    vid = int(vid, 16)
                    pid = int(pid, 16)
                except ValueError as e:
                    print(f"Invalid vid/pid in {match} in {f}", file=sys.stderr)
                    raise e

                if bus == "usb":
                    bus = 0x3
                elif bus == "bluetooth":
                    bus = 0x5

                class Tablet(object):
                    def __init__(self, name, bus, vid, pid, is_touchscreen=False):
                        self.name = name
                        self.bus = bus
                        self.vid = vid
                        self.pid = pid
                        self.is_touchscreen = is_touchscreen

                tablets.append(Tablet(name, bus, vid, pid))

                if want_pad:
                    tablets.append(Tablet(name + " Pad", bus, vid, pid))

                if want_finger:
                    tablets.append(
                        Tablet(name + " Finger", bus, vid, pid, is_touchscreen)
                    )

        # our tablets list now becomes the list of arguments passed to the
        # test functions taking a 'tablet' argument - one-by-one. So where
        # tablets contains 10 entries, our test function will be called 10
        # times.
        metafunc.parametrize("tablet", tablets, ids=[t.name for t in tablets])


@pytest.mark.skipif(sys.platform != "linux", reason="This test requires udev")
def test_hwdb_files(tablet):
    # Note: the actual name doesn't really matter, all our hwdb files use either "*"
    # or "* Finger", etc. It does matter for that "Finger" suffix though.
    query = f"libwacom:name:{tablet.name}:input:b{tablet.bus:04X}v{tablet.vid:04X}p{tablet.pid:04X}"
    logging.debug(query)

    r = subprocess.run(
        ["systemd-hwdb", "query", query], check=True, capture_output=True
    )
    stdout = r.stdout.decode("utf-8").strip()
    assert stdout, f"No output recorded for query {query}"
    logging.debug(stdout)
    props = {}
    for line in filter(lambda line: len(line) > 1, stdout.split("\n")):
        print(line)
        k, v = line.split("=")
        props[k] = v

    assert "ID_INPUT" in props
    assert props["ID_INPUT"] == "1"

    assert "ID_INPUT_TABLET" in props
    assert props["ID_INPUT_TABLET"] == "1"

    if "ID_INPUT_JOYSTICK" not in props:
        assert props["ID_INPUT_JOYSTICK"] == "0"

    if "Finger" in tablet.name:
        if tablet.is_touchscreen:
            assert "ID_INPUT_TOUCHSCREEN" in props
        else:
            assert "ID_INPUT_TOUCHPAD" in props

    # For the Wacom Bamboo Pad we check for "Pad Pad" in the device name
    if "Pad" in tablet.name:
        if "Wacom Bamboo Pad" not in tablet.name or "Pad Pad" in tablet.name:
            assert "ID_INPUT_TABLET_PAD" in props



#!/usr/bin/env python3
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without
# fee, provided that the above copyright notice appear in all copies
# and that both that copyright notice and this permission notice
# appear in supporting documentation, and that the name of Red Hat
# not be used in advertising or publicity pertaining to distribution
# of the software without specific, written prior permission.  Red
# Hat makes no representations about the suitability of this software
# for any purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import argparse
import configparser
import os
import sys
from pathlib import Path

try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(str(e)), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)


def xdg_dir():
    return Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config")) / "libwacom"


class Ansi:
    clearline = "\x1b[K"

    @classmethod
    def up(cls, count):
        return f"\x1b[{count}A"

    @classmethod
    def down(cls, count):
        return f"\x1b[{count}B"

    @classmethod
    def right(cls, count):
        return f"\x1b[{count}C"

    @classmethod
    def left(cls, count):
        return f"\x1b[{count}D"


def die(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def select_device():
    context = pyudev.Context()
    for device in context.list_devices(subsystem="input"):
        if device.get("ID_INPUT_TABLET", 0) and (device.device_node or "").startswith(
            "/dev/input/event"
        ):
            name = device.get("NAME", None)
            if not name:
                name = next(
                    (p.get("NAME") for p in device.ancestors if p.get("NAME")),
                    "unknown",
                )

            print("Using {}: {}".format(name or "unknown", device.device_node))
            return device.device_node

    die("Unable to find a tablet device.")


def record_events(ns):
    with open(ns.device_path, "rb") as fd:
        d = libevdev.Device(fd)
        if not d.absinfo[libevdev.EV_ABS.ABS_MISC]:
            die("Device only supports generic styli")

        tool_bits = set(
            c for c in libevdev.EV_KEY.codes if c.name.startswith("BTN_TOOL_")
        )
        styli = {}  # dict of (type, serial) = proximity_state
        current_type, current_serial = 0, 0
        in_prox = False
        dirty = False

        print("Please put tool in proximity")

        try:
            while True:
                for event in d.events():
                    if event.matches(libevdev.EV_ABS.ABS_MISC):
                        if event.value != 0:
                            current_type = event.value
                            dirty = True
                    elif event.matches(libevdev.EV_MSC.MSC_SERIAL):
                        if event.value != 0:
                            current_serial = event.value & 0xFFFFFFFF
                            dirty = True
                    elif event.code in tool_bits:
                        # print(f'Current prox: {event.value}')
                        in_prox = event.value != 0
                        dirty = True
                    elif event.matches(libevdev.EV_SYN.SYN_REPORT) and dirty:
                        dirty = False
                        print(
                            f"{Ansi.up(len(styli))}{Ansi.left(10000)}{Ansi.clearline}",
                            end="",
                        )
                        styli[(current_type, current_serial)] = in_prox
                        for s, prox in styli.items():
                            tid, serial = s
                            print(
                                f"Tool id {tid:#x} serial {serial:#x} in-proximity: {prox} "
                            )
        except KeyboardInterrupt:
            print("Terminating")

        return [s[0] for s in styli.keys()]


def load_data_files():
    lookup_paths = (
        ("./data/",),
        ("@DATADIR@", "@ETCDIR@", xdg_dir()),
        ("/usr/share/libwacom/", "/etc/libwacom/"),
    )
    stylusfiles = []
    for paths in lookup_paths:
        stylusfiles = []
        for p in paths:
            files = list(Path(p).glob("*.stylus"))
            if files:
                stylusfiles += files

        if any(stylusfiles):
            break
    else:
        die("Unable to find a libwacom.stylus data file")

    print(f"Using stylus file(s): {', '.join([str(s) for s in stylusfiles])}")

    styli = {}

    for path in stylusfiles:
        config = configparser.ConfigParser()
        config.read(path)
        for stylus_id in config.sections():
            ids = stylus_id.split(":")
            if len(ids) > 1:
                _, sid = map(lambda x: int(x, 16), ids)
            else:
                _ = 0x56A  # vid
                sid = int(ids[0], 16)
            # FIXME: vendor should be used here, let's do that when we figure out
            # who needs it.
            styli[sid] = config[stylus_id].get("Group", stylus_id)

    return styli


def main():
    parser = argparse.ArgumentParser(description="Tool to show tablet stylus ids")
    parser.add_argument(
        "device_path", nargs="?", default=None, help="Path to the /dev/input/event node"
    )

    ns = parser.parse_args()
    if not ns.device_path:
        ns.device_path = select_device()

    all_styli = load_data_files()
    styli = record_events(ns)
    groups = []
    for sid in styli:
        if sid in all_styli:
            groups.append(all_styli[sid])
        else:
            print(f"Unknown stylus id {sid:#x}. New entry needed")
    print("Suggested line for .tablet file:")
    print(f"Styli={';'.join(set(groups))}")


if __name__ == "__main__":
    try:
        main()
    except PermissionError:
        die("Insufficient permissions, please run me as root")



#!/usr/bin/env python3
#
# This file is formatted with ruff format
#

from pathlib import Path
import logging
import os
import pytest

from . import WacomDatabase

logger = logging.getLogger(__name__)


def load_test_db() -> WacomDatabase:
    try:
        dbpath = os.environ.get("MESON_SOURCE_ROOT")
        if dbpath is None and (Path.cwd() / "meson.build").exists():
            dbpath = Path.cwd()
            logger.info(f"Defaulting to MESON_SOURCE_ROOT={dbpath}")
        return WacomDatabase(path=Path(dbpath) / "data" if dbpath else None)
    except AttributeError as e:
        pytest.exit(
            f"Failed to initialize and wrap libwacom.so: {e}. You may need to set LD_LIBRARY_PATH and optionally MESON_SOURCE_ROOT"
        )


@pytest.fixture()
def db():
    return load_test_db()



#!/usr/bin/env python3
#
# Run with pytest

import configparser
import os
import re
from pathlib import Path


WACOM_RECEIVER_USBIDS = [
    (0x56A, 0x84),
]


def datadir():
    return Path(os.getenv("MESON_SOURCE_ROOT") or ".") / "data"


def layoutsdir():
    return datadir() / "layouts"


def pytest_generate_tests(metafunc):
    # for any function that takes a "tabletfile" argument return the path to
    # a tablet file
    if "tabletfile" in metafunc.fixturenames:
        files = [f for f in datadir().glob("*.tablet")]
        metafunc.parametrize("tabletfile", files, ids=[f.name for f in files])


def test_device_match(tabletfile):
    config = configparser.ConfigParser()
    config.read(tabletfile)

    # Match format must be bus:vid:pid:name
    # where bus is 'usb' or 'bluetooth'
    # where vid/pid is a lowercase hex
    # where name is optional
    for match in config["Device"]["DeviceMatch"].split(";"):
        if not match or match == "generic":
            continue

        bus, vid, pid = match.split("|")[:3]  # skip the name part of the match
        assert bus in [
            "usb",
            "bluetooth",
            "i2c",
            "serial",
        ], f"{tabletfile}: unknown bus type"
        assert re.match("[0-9a-f]{4}", vid), (
            f"{tabletfile}: {vid} must be lowercase hex"
        )
        assert re.match("[0-9a-f]{4}", pid), (
            f"{tabletfile}: {pid} must be lowercase hex"
        )


def test_no_receiver_id(tabletfile):
    config = configparser.ConfigParser(strict=True)
    # Don't convert to lowercase
    config.optionxform = lambda option: option
    config.read(tabletfile)

    receivers = ["usb|{:04x}|{:04x}".format(*r) for r in WACOM_RECEIVER_USBIDS]
    for match in config["Device"]["DeviceMatch"].split(";"):
        assert match not in receivers


def test_svg_exists(tabletfile):
    config = configparser.ConfigParser(strict=True)
    # Don't convert to lowercase
    config.optionxform = lambda option: option
    config.read(tabletfile)

    try:
        svg = config["Device"]["Layout"]
        assert svg != "", "Empty Layout= line not permitted"
        assert (layoutsdir() / svg).exists()

    except KeyError:
        pass


def test_button_evcodes(tabletfile):
    config = configparser.ConfigParser(strict=True)
    # Don't convert to lowercase
    config.optionxform = lambda option: option
    config.read(tabletfile)

    try:
        nbuttons = 0
        for where in ["Top", "Bottom", "Left", "Right"]:
            try:
                buttons = config["Buttons"][where]
                nbuttons += len([s for s in buttons.split(";") if s])
            except KeyError:
                pass
        str = config["Buttons"]["EvdevCodes"]
        codes = [
            c for c in str.split(";") if c
        ]  # drop empty strings from trailing semicolons
        assert len(codes) == nbuttons, "Number of buttons mismatches the EvdevCodes"
    except KeyError:
        pass
